import pandas as pd

pd.options.display.float_format = "{:12.5e}".format


def _validate_input(association: str, max_extension: int, basal_upstream: int, basal_downstream: int) -> bool:
    """
    Checks that the inputs (association_rule, max_extension, basal_upstream, basal_downstream) are valid

    Parameters
    ----------
    association : str
        The association rule to use. Documentation aviable at https://great-help.atlassian.net/wiki/spaces/GREAT/pages/655443/Association+Rules\n
    max_extension : int
        The maximum extension of the regulatory domain.\n
    basal_upstream : int
        The basal upstream of the regulatory domain.\n
    basal_downstream : int
        The basal downstream of the regulatory domain.\n

    Returns
    -------
    bool
        True if the inputs are valid, False otherwise.

    Examples
    --------
    >>> validate_input("two_closet")
    ...    True

    >>> validate_input("Two_Closet_Extension")
    ...    False
    """
    if association != "one_closet" and association != "two_closet" and association != "basal_plus_extention":
        print("Association rule should be OneCloset or TwoCloset Basalplusextention")
        return False
    if max_extension < 0:
        print(f"Maximum extension must be a non-negative integer: {max_extension}")
        return False
    if basal_upstream < 0:
        print(f"Basal upstream must be a non-negative integer: {basal_upstream}")
        return False
    if basal_downstream < 0:
        print(f"Basal downstream must be a non-negative integer: {basal_downstream}")
        return False
    return True


def _write_regdom(regdom: pd.DataFrame, file_name: str) -> None:
    """
    Write the regulatory regions calculated in a file given in argument

    Parameters
    ----------
    regdom : pd.DataFrame
        The regulatory regions to write.\n
    file_name : str
        The path of the file to write.\n

    Returns
    -------
    None.
    """
    f = open(file_name, "w")
    f.write("#chr\tchrstart\tchrend\tname\ttss\tstrand\n")
    for i in range(regdom.shape[0]):
        curr = regdom.iloc[i]
        chr = curr["chr"]
        start = curr["chr_start"]
        end = curr["chr_end"]
        name = curr["name"]
        tss = curr["tss"]
        strand = curr["strand"]
        f.write(f"{chr}\t{start}\t{end}\t{name}\t{tss}\t{strand}\n")
    f.close()


def _create_basal_plus_extension_regdom(
    tss: pd.DataFrame, maximumExtension: int, basalUp: int, basalDown: int, chr_size: pd.DataFrame
) -> pd.DataFrame:
    """
    Create the regulatory domains using the Basalplusextention association rule

    Parameters
    ----------
    tss : pd.DataFrame
        The TSS of the genes.\n
    maximumExtension : int
        The maximum extension of the regulatory domain.\n
    basalUp : int
        The basal upstream of the regulatory domain.\n
    basalDown : int
        The basal downstream of the regulatory domain.\n
    chr_size : pd.DataFrame
        The chromosome size.\n

    Returns
    -------
    tss : pd.DataFrame
        The regulatory domains.

    Examples
    --------
    >>> regdom = create_basal_plus_extension_regdom(
        tss_file=pd.read_csv("../../data/human/tss.bed",sep="\t",names=["chr","tss","strand"]),
        maximumExtension=100000,
        basalUp=5000,
        basalDown=1000,
        chr_sizes=pd.read_csv("../../data/human/chr_size.bed",sep="\t",names=["chr","size"])
    )
    >>> regdom.head()
    ...    |    | chr   |   chr_start |   chr_end | name      |   tss | strand   |
    ...    |---:|:------|------------:|----------:|:----------|------:|:---------|
    ...    |  0 | chr1  |           0 |     22436 | MIR6859-1 | 17436 | -        |
    ...    |  1 | chr1  |       16436 |     22436 | MIR6859-2 | 17436 | -        |
    ...    |  2 | chr1  |       16436 |     22436 | MIR6859-3 | 17436 | -        |
    ...    |  3 | chr1  |       16436 |     28370 | MIR6859-4 | 17436 | -        |
    ...    |  4 | chr1  |       22436 |     34370 | WASH7P    | 29370 | -        |

    """
    prev = 0
    curr = 0
    next = 0
    chr_strat_end = []
    start = []
    end = []
    for i in range(tss.shape[0]):
        curr = tss.iloc[i]
        chr = curr["chr"]
        curr_chr_size = int(chr_size.loc[chr_size["chr"] == chr, "size"])
        tmp = curr["tss"]
        if curr["strand"] == "+":
            curr_chr_start = max(0, tmp - basalUp)
            curr_chr_end = min(curr_chr_size, tmp + basalDown)
        elif curr["strand"] == "-":
            curr_chr_start = max(0, tmp - basalDown)
            curr_chr_end = min(curr_chr_size, tmp + basalUp)
        elif curr["strand"] == ".":
            print("Invalid_Input : Impossible to create a basal expression regdom if you have not specify the strand")
            return False
        else:
            err = curr["strand"]
            print(f"Invalid input : strand should be '+' or '-'. Line {i} : strand = {err}")
            return False
        chr_strat_end.append([curr_chr_start, curr_chr_end])

    for i in range(tss.shape[0]):
        curr = tss.iloc[i]
        chr = curr["chr"]

        curr_chr_size = int(chr_size.loc[chr_size["chr"] == chr, "size"])
        if i != tss.shape[0] - 1:
            next = tss.iloc[i + 1]
        else:
            next = 0

        tmp_start = max(0, curr["tss"] - maximumExtension)
        basal_start = chr_strat_end[i][0]
        tmp_start = min(basal_start, tmp_start)
        if type(prev) != int and prev["chr"] == curr["chr"]:
            if prev["strand"] == "+":
                prev_end = prev["tss"] + basalDown
            else:
                prev_end = prev["tss"] + basalUp
            tmp_start = min(basal_start, max(prev_end, tmp_start))

        tmp_end = min(curr_chr_size, curr["tss"] + 1000000)
        basal_end = chr_strat_end[i][1]

        tmp_end = max(basal_end, tmp_end)
        if type(next) != int and next["chr"] == curr["chr"]:
            if next["strand"] == "+":
                nextstart = next["tss"] - basalUp
            else:
                nextstart = next["tss"] - basalDown
            tmp_end = max(basal_end, min(nextstart, tmp_end))

        prev = tss.iloc[i]
        start.append(int(tmp_start))
        end.append(int(tmp_end))
    tss["chr_start"] = start
    tss["chr_end"] = end
    return tss


def _create_two_closet_regdom(tss: pd.DataFrame, max_extension: int, chr_size: pd.DataFrame) -> pd.DataFrame:
    """
    Create the regulatory domains using the TwoCloset association rule.\n
    It is based on the basal plus extension rule but with basalUp and basalDown equals to 0.\n

    Parameters
    ----------
    tss : pd.DataFrame
        The TSS of the genes.\n
    maximumExtension : int
        The maximum extension of the regulatory domain.\n
    chr_size : pd.DataFrame
        The chromosome size.\n

    Returns
    -------
    pd.DataFrame
        The regulatory domains.

    Examples
    --------
    >>> regdom = _create_two_closet_regdom(
        tss_file=pd.read_csv("../../data/human/tss.bed",sep="\t",names=["chr","tss","strand"]),
        maximumExtension=100000,
        chr_sizes=pd.read_csv("../../data/human/chr_size.bed",sep="\t",names=["chr","size"])
    )
    >>> regdom.head()
    ...    |    | chr   |   chr_start |   chr_end | name      |   tss | strand   |
    ...    |---:|:------|------------:|----------:|:----------|------:|:---------|
    ...    |  0 | chr1  |           0 |     17436 | MIR6859-1 | 17436 | -        |
    ...    |  1 | chr1  |       17436 |     17436 | MIR6859-2 | 17436 | -        |
    ...    |  2 | chr1  |       17436 |     17436 | MIR6859-3 | 17436 | -        |
    ...    |  3 | chr1  |       17436 |     29370 | MIR6859-4 | 17436 | -        |
    ...    |  4 | chr1  |       17436 |     30365 | WASH7P    | 29370 | -        |

    """
    return _create_basal_plus_extension_regdom(tss, max_extension, 0, 0, chr_size)


def _create_one_closet_regdom(tss: pd.DataFrame, maximum_extension: int, chr_size: pd.DataFrame) -> pd.DataFrame:
    """
    Create the regulatory domains using the OneCloset association rule

    Parameters
    ----------
    tss : pd.DataFrame
        The TSS of the genes.\n
    maximumExtension : int
        The maximum extension of the regulatory domain.\n
    basalUp : int
        The basal upstream of the regulatory domain.\n
    basalDown : int
        The basal downstream of the regulatory domain.\n
    chr_size : pd.DataFrame
        The chromosome size.\n

    Returns
    -------
    tss : pd.DataFrame
        The regulatory domains.

    Examples
    --------
    >>> regdom = create_basal_plus_extension_regdom(
        tss_file=pd.read_csv("../../data/human/tss.bed",sep="\t",names=["chr","tss","strand"]),
        maximum_extension=100000,
        chr_sizes=pd.read_csv("../../data/human/chr_size.bed",sep="\t",names=["chr","size"])
    )
    >>> regdom.head()
    ...    |    | chr   |   chr_start |   chr_end | name      |   tss | strand   |
    ...    |---:|:------|------------:|----------:|:----------|------:|:---------|
    ...    |  0 | chr1  |           0 |     17436 | MIR6859-1 | 17436 | -        |
    ...    |  1 | chr1  |       17436 |     17436 | MIR6859-2 | 17436 | -        |
    ...    |  2 | chr1  |       17436 |     17436 | MIR6859-3 | 17436 | -        |
    ...    |  3 | chr1  |       17436 |     23403 | MIR6859-4 | 17436 | -        |
    ...    |  4 | chr1  |       23403 |     29867 | WASH7P    | 29370 | -        |

    """
    prev = 0
    curr = 0
    next = 0
    start = []
    end = []

    for i in range(tss.shape[0]):
        curr = tss.iloc[i]
        chr = curr["chr"]
        curr_chr_size = int(chr_size.loc[chr_size["chr"] == chr, "size"])
        if i < tss.shape[0] - 1:
            next = tss.iloc[i + 1]
        else:
            next = 0

        if (type(prev) == int and prev == 0) or (type(prev) != int and prev["chr"] != curr["chr"]):
            prev = 0
        if (type(next) == int and next == 0) or (type(next) != int and next["chr"] != curr["chr"]):
            next = 0

        tmp_start = max(0, curr["tss"] - maximum_extension)
        if type(prev) != int:
            middle = (curr["tss"] + prev["tss"]) // 2
            tmp_start = max(tmp_start, middle)

        tmp_end = min(curr["tss"] + maximum_extension, curr_chr_size)
        if type(next) != int:
            middle = (curr["tss"] + next["tss"]) // 2
            tmp_end = min(tmp_end, middle)

        start.append(tmp_start)
        end.append(tmp_end)
        prev = curr

    tss["chr_start"] = start
    tss["chr_end"] = end
    return tss


def create_regdom(
    tss_file: str,
    chr_sizes_file: str,
    association_rule: str,
    max_extension: int = 1000000,
    basal_upstream: int = 5000,
    basal_downstream: int = 1000,
    out_path: str or None = None,
) -> pd.DataFrame:
    """\
    Create regdoms according to the three association rules, to write the result in a file or not and to return the result as a pd.DataFrame

    Parameters
    ----------
    tss_file : str
        The path of the TSS file.\n
    chr_sizes_file : str
        The path of the chromosome size file.\n
    association_rule : str
        The association rule to use. Could be : "one_closet", "two_closet", "basal_plus_extention".\n
        Documentation aviable at https://great-help.atlassian.net/wiki/spaces/GREAT/pages/655443/Association+Rules.\n
    maximumExtension : int
        The maximum extension of the regulatory domain.\n
        Default is `100000`\n
    basalUp : int
        The basal upstream of the regulatory domain.\n
        Default is `5000`\n
    basalDown : int
        The basal downstream of the regulatory domain.\n
        Default is `1000`\n
    out_path : str or NoneType
        The path of the output file.\n
        If None, the result is only returned as a pd.DataFrame.\n
        Default is `None`\n

    Returns
    -------
    out : pd.DataFrame
        The regulatory domains.

    Examples
    --------
    >>> regdom = create_regdom(
        tss_file="../../data/human/tss.bed",
        chr_sizes_file="../../data/human/chr_size.bed",
        sep="\t",
        names=["chr","size"],
        association_rule="one_closet"
        )
    >>> regdom.head()
    ...    |    | chr   |   chr_start |   chr_end | name      |   tss | strand   |
    ...    |---:|:------|------------:|----------:|:----------|------:|:---------|
    ...    |  0 | chr1  |           0 |     17436 | MIR6859-1 | 17436 | -        |
    ...    |  1 | chr1  |       17436 |     17436 | MIR6859-2 | 17436 | -        |
    ...    |  2 | chr1  |       17436 |     17436 | MIR6859-3 | 17436 | -        |
    ...    |  3 | chr1  |       17436 |     23403 | MIR6859-4 | 17436 | -        |
    ...    |  4 | chr1  |       23403 |     29867 | WASH7P    | 29370 | -        |

    """
    if not _validate_input(association_rule, max_extension, basal_upstream, basal_downstream):
        print("Invalid input")
        return False
    df = pd.read_csv(tss_file, sep="\t", comment="#", names=["chr", "tss", "strand", "name"])

    df = df.sort_values(["chr", "tss", "strand", "name"])

    chr_size = pd.read_csv(chr_sizes_file, sep="\t", comment="#", names=["chr", "size"])

    if association_rule == "one_closet":
        out = _create_one_closet_regdom(df, max_extension, chr_size)
    elif association_rule == "two_closet":
        out = _create_two_closet_regdom(df, max_extension, chr_size)
    elif association_rule == "basal_plus_extention":
        out = _create_basal_plus_extension_regdom(df, max_extension, basal_upstream, basal_downstream, chr_size)
    else:
        return False
    out = out.astype({"chr_start": int, "chr_end": int})
    out = out.reindex(["chr", "chr_start", "chr_end", "name", "tss", "strand"], axis=1)

    if out_path is not None:
        _write_regdom(out, out_path)
    return out



import re

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import rpy2
import seaborn as sns
from rpy2.robjects import pandas2ri
from rpy2.robjects import r as r
from rpy2.robjects.packages import importr
from scipy.stats import pearsonr

import greatpy as great

pandas2ri.activate()


def online_vs_local_vs_greatpy_comparison(
    input_folder: str = "../data/tests/test_data/input/",
    information_folder: str = "../data/human/",
    annotation_file: str = "../../data/human/ontologies.csv",
):
    """
    Make a comparison between the online and the local version of rGREAT and greatpy.\n
    The function return a clustermap of the results between online vs local and greatpy.\n

    Parameters
    ----------
    input_folder : str
        path of the folder with the input files for the tests.\n
        Default is `"../data/tests/test_data/input/"`. \n
    information_folder : str
        path of the folder with the information files for the tests. \n
        Default is `"../data/human/"`\n
        The input folder should contains the files : \n
        - information_folder/assembly_eg_hg38/regulatory_domain.bed \n
        - information_folder/assembly_eg_hg38/chr_size.bed\n
    annotation_file : str
        path of the file with the ontologies annotations. The file should be in the csv format with `;` as separator and contain the following columns :\n
        - "ensembl": ensemble id (optionnal)\n
        - "id": id of the ontology\n
        - "name": name of the ontology\n
        - "ontology.group": group of the ontology\n
        - "gene.name": name of the gene\n
        - "symbol": symbol of the gene equivalent to the gene.name\n
        Default is `"../../data/human/ontologies.csv"`\n

    Returns
    -------
    None

    Note
    ----
    To use this function, you should have installed in your environment:
        - rpy2
        - R base with the version 3.6.1
        - The following R packages : `rGREAT`, `GenomicRanges`

    """
    importr("rGREAT")
    ranges = importr("GenomicRanges")

    ann = pd.read_csv(
        annotation_file,
        sep=";",
        comment="#",
        header=0,
    )
    ann["name"] = ann["name"].str.lower()
    pd.DataFrame()

    res = {"name1": [], "file": [], "name3": [], "value": []}

    for name in [
        "01_random.bed",
        "04_ultra_hg38.bed",
        "06_height_snps_hg38.bed",
        "07_height_snps_hg19.bed",
        "10_MAX.bed",
    ]:
        # find the assembly
        if re.match(".*hg19.*", name) is not None:
            assembly = "hg19"
        else:
            assembly = "hg38"

        # online test
        res_online = rpy2.robjects.r["submitGreatJob"](f"{input_folder}{name}", species=f"{assembly}", help=False)
        res_online = rpy2.robjects.r["getEnrichmentTables"](res_online)

        # local test
        # proprocessing : make a Grange frame
        df = r["read.csv"](f"{input_folder}{name}", sep="\t")
        seqname = rpy2.robjects.StrVector(
            ["seqnames", "seqname", "chromosome", "X.Chr", "chr", "chromosome_name", "seqid"]
        )
        rpy2.robjects.StrVector(["end", "stop"])
        df = ranges.makeGRangesFromDataFrame(df, seqnames_field=seqname)

        # great calculation
        local = rpy2.robjects.r["great"](df, "GREAT:C5", f"txdb:{assembly}", verbose=False)
        local = rpy2.robjects.r["getEnrichmentTables"](local)

        # greatpy calculation
        greatpy = great.tl.enrichment(
            test_file=f"{input_folder}{name}",
            regdom_file=f"{information_folder}{assembly}/regulatory_domain.bed",
            chr_size_file=f"{information_folder}{assembly}/chr_size.bed",
            annotation_file=f"{annotation_file}",
            binom=True,
            hypergeom=True,
        )

        # create each dataframe
        # online
        name_online = [
            cdc.lower()
            for cdc in list(res_online.rx2("GO Molecular Function").rx2("name"))
            + list(res_online.rx2("GO Biological Process").rx2("name"))
            + list(res_online.rx2("GO Cellular Component").rx2("name"))
        ]
        online = pd.DataFrame(
            {
                "id": list(res_online.rx2("GO Molecular Function").rx2("ID"))
                + list(res_online.rx2("GO Biological Process").rx2("ID"))
                + list(res_online.rx2("GO Cellular Component").rx2("ID")),
                "name": name_online,
                "binom_p_val": list(res_online.rx2("GO Molecular Function").rx2("Binom_Raw_PValue"))
                + list(res_online.rx2("GO Biological Process").rx2("Binom_Raw_PValue"))
                + list(res_online.rx2("GO Cellular Component").rx2("Binom_Raw_PValue")),
                "hyper_p_val": list(res_online.rx2("GO Molecular Function").rx2("Hyper_Raw_PValue"))
                + list(res_online.rx2("GO Biological Process").rx2("Hyper_Raw_PValue"))
                + list(res_online.rx2("GO Cellular Component").rx2("Hyper_Raw_PValue")),
            }
        )

        # local
        name_local = list(local.rx2("id"))
        name_local = [" ".join(cdc.lower().split("_")[1:]) for cdc in list(local.rx2("id"))]
        local = pd.DataFrame(
            {
                "name": name_local,
                "binom_p_val": list(local.rx2("p_value")),
                "hyper_p_val": list(local.rx2("p_value_hyper")),
            }
        )

        # greatpy
        greatpy["go_term"] = greatpy["go_term"].str.lower()

        # correlation between online and greatpy
        o_great = online[online["id"].isin(list(greatpy.index))]
        greatpy[greatpy.index.isin(list(online["id"]))]

        tot = pd.DataFrame()
        res_bin = []
        res_hyp = []
        cols_to_supr = []
        for i in range(len(o_great)):
            curr_o = o_great.iloc[i]
            if greatpy.loc[greatpy.index == curr_o["id"]].shape[0] > 0:
                res_bin.append(float(greatpy.loc[greatpy.index == curr_o["id"]]["binom_p_value"]))
                res_hyp.append(float(greatpy.loc[greatpy.index == curr_o["id"]]["hypergeom_p_value"]))
            else:
                cols_to_supr.append(i)

        tot["id"] = o_great["id"]
        tot["binom_p_value_online"] = -np.log(list(o_great["binom_p_val"]))
        tot["hyper_p_value_online"] = -np.log(list(o_great["hyper_p_val"]))
        tot["binom_p_value_greatpy"] = -np.log(res_bin)
        tot["hypergeom_p_value_greatpy"] = -np.log(res_hyp)

        # add GO id to local
        go = []
        for i in range(local.shape[0]):
            curr = local.iloc[i]

            if ann.loc[ann["name"].isin([curr["name"]])].shape[0] > 0:
                go.append(ann.loc[ann["name"].isin([curr["name"]])].iloc[0]["id"])
            else:
                go.append("")
        local["id"] = go
        local = local.loc[local["id"] != ""]

        # Correlation between online and local
        o_loc = online[online["id"].isin(local["id"])]
        loc_o = local[local["id"].isin(online["id"])]
        tot_2 = pd.DataFrame()

        res_bin = []
        res_hyp = []
        cols_to_supr = []
        for i in range(len(o_loc)):
            curr_o = o_loc.iloc[i]
            if loc_o.loc[loc_o["id"] == curr_o["id"]].shape[0] > 0:
                res_bin.append(float(loc_o.loc[loc_o["id"] == curr_o["id"]].iloc[0]["binom_p_val"]))
                res_hyp.append(float(loc_o.loc[loc_o["id"] == curr_o["id"]].iloc[0]["hyper_p_val"]))
            else:
                cols_to_supr.append(i)
        tot_2["id"] = o_loc["id"]
        tot_2["binom_p_value_online"] = -np.log(list(o_loc["binom_p_val"]))
        tot_2["hyper_p_value_online"] = -np.log(list(o_loc["hyper_p_val"]))
        tot_2["binom_p_value_local"] = -np.log(res_bin)
        tot_2["hypergeom_p_value_local"] = -np.log(res_hyp)

        # Correlation between greatpy and local
        g_loc = greatpy[greatpy.index.isin(local["id"])]
        loc_g = local[local["id"].isin(greatpy.index)]
        tot_3 = pd.DataFrame()

        res_bin = []
        res_hyp = []
        cols_to_supr = []
        g_loc = g_loc.reset_index().rename(columns={"index": "id"})
        for i in range(g_loc.shape[0]):
            curr_o = g_loc.iloc[i]
            if loc_g.loc[loc_g["id"] == curr_o["id"]].shape[0] > 0:
                res_bin.append(float(loc_g.loc[loc_g["id"] == curr_o["id"]].iloc[0]["binom_p_val"]))
                res_hyp.append(float(loc_g.loc[loc_g["id"] == curr_o["id"]].iloc[0]["hyper_p_val"]))
            else:
                cols_to_supr.append(i)
        tot_3["id"] = g_loc["id"]
        tot_3["binom_p_value_greatpy"] = -np.log(list(g_loc["binom_p_value"]))
        tot_3["hyper_p_value_greatpy"] = -np.log(list(g_loc["hypergeom_p_value"]))
        tot_3["binom_p_value_local"] = -np.log(res_bin)
        tot_3["hypergeom_p_value_local"] = -np.log(res_hyp)

        tot = tot.replace(np.inf, np.nan)
        tot_2 = tot_2.replace(np.inf, np.nan)
        tot_3 = tot_3.replace(np.inf, np.nan)

        tot = tot.dropna()
        tot_2 = tot_2.dropna()
        tot_3 = tot_3.dropna()

        # create the table
        name_plot = " ".join(name.split(".")[0].split("_")[1:])
        res["name1"].append("online")
        res["file"].append(name_plot)
        res["name3"].append("greatpy")
        res["value"].append(pearsonr(tot["binom_p_value_online"], tot["binom_p_value_greatpy"])[0])

        res["name1"].append("online")
        res["file"].append(name_plot)
        res["name3"].append("local")
        res["value"].append(pearsonr(tot_2["binom_p_value_online"], tot_2["binom_p_value_local"])[0])

    res = pd.DataFrame(res)
    res = res.pivot("file", "name3", "value")
    g = sns.heatmap(data=res, cmap="Reds", annot=True)
    g.set_title("correlation with Great server")
    g.set_ylabel("online results for each file")
    g.set_xlabel("algorithm results for each file")

    plt.show(g)



from .basic import (
    dotplot_multi_sample,
    get_all_comparison,
    graph_absolute_dist_tss,
    graph_dist_tss,
    graph_nb_asso_per_peaks,
    make_bubble_heatmap,
    plot_enrich,
    scatterplot,
)



import os
import re

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import colors, rc, rcParams
from numpy import log
from scipy.stats import pearsonr

import greatpy as great

plt.rcParams.update({"font.size": 14, "font.weight": "normal"})


def scatterplot(
    great_df: pd.DataFrame,
    colname_x: str,
    colname_y: str,
    title: str = "",
    minus_log10=True,
    ax: matplotlib.axes or None = None,
) -> None:
    """
    Create a scatterplot from a
    pandas dataframe between two columns.\n
    A logarithmic scale can be used.\n

    Parameters
    ----------
    great_df : pd.DataFrame
        Output of the greatpy.tl.enrichment function\n
    colname_x : str
        Name of the column to be used as x axis\n
    colname_y : str
        Name of the column to be used as y axis\n
    title : str
        Title of the plot.
        Default is `""`\n
    minus_log10 : bool
        If True, the logarithmic scale is used\n
        Default is `True`\n
    ax : matplotlib.axes or None
        Define the position of the plot in a figure\n
        Default is `None`\n

    Returns
    -------
    None
        Scatterplot between the two columns passed in args

    """
    great_df = great_df.dropna()
    great_df = great_df.astype({colname_x: "float64", colname_y: "float64"})
    if minus_log10:
        great_df[f"-log({colname_x})"] = -log(great_df[colname_x])
        great_df[f"-log({colname_y})"] = -log(great_df[colname_y])
        g = sns.scatterplot(data=great_df, x=f"-log({colname_x})", y=f"-log({colname_y})", ax=ax)
        g.set_title(title, fontsize=18)
        g.set_xlabel(f"-log({colname_x})", fontsize=12)
        g.set_ylabel(f"-log({colname_y})", fontsize=12)
    else:
        g = sns.scatterplot(data=great_df, x=colname_x, y=colname_y, ax=ax)
        g.set_title(title, fontsize=20)
        g.set_xlabel(colname_x, fontsize=15)
        g.set_ylabel(colname_y, fontsize=15)


def graph_nb_asso_per_peaks(
    test: str or pd.DataFrame,
    regdom: str or pd.DataFrame,
    ax: matplotlib.axes or None = None,
    color: str = None,
) -> None:
    """
    Creates a barplot representing the
    percentage of peaks for all possible association numbers

    Parameters
    ----------
    test : str or pd.DataFrame
        Genomic set of peaks to be tested\n
    regdom : str or pd.DataFrame
        Regulatory domain of all genes in the genome\n
    ax : matplotlib.axes or None
        Define the position of the plot in a figure\n
        Default is `None`\n
    color : str
        Color of the bar\n
        Default is `None` \n

    Returns
    -------
    None
        Barplot of the number of associated genes per peak

    """
    nb_asso_per_peaks = great.tl.get_nb_asso_per_region(test, regdom)

    nb = {
        "number": [],
        "number_genes": [],
        "percentage": [],
    }
    for i in list(set(nb_asso_per_peaks.values())):
        nb["number"].append(i)
        nb["number_genes"].append(list(nb_asso_per_peaks.values()).count(i))
        nb["percentage"].append(
            round((list(nb_asso_per_peaks.values()).count(i) / len(nb_asso_per_peaks.keys())) * 100)
        )
    nb = pd.DataFrame(nb, columns=["number", "number_genes", "percentage"], index=nb["number"])

    g = sns.barplot(data=nb, x="number", y="percentage", ax=ax, color=color)
    g.set_title("Number of associated genes per region", fontsize=20)
    g.set_xlabel("Number of associated genes per region", fontsize=15)
    g.set_ylabel("Genomic region (%)", fontsize=15)

    for i in range(nb.shape[0]):
        x = nb.iloc[i]["number"]
        y = nb.iloc[i]["percentage"]
        g.text(x=x - 0.06, y=y + 1, s=nb.number_genes[i])


def graph_dist_tss(
    test: str or pd.DataFrame,
    regdom: str or pd.DataFrame,
    ax: matplotlib.axes or None = None,
    color: str = "#325fa8",
) -> None:
    """
    Creation of a barplot of the distance
    between the peaks and the TSS of the associated gene(s).\n

    Parameters
    ----------
    test : str or pd.DataFrame
        Genomic set of peaks to be tested\n
    regdom : str or pd.DataFrame
        Regulatory domain of all genes in the genome\n
    ax : matplotlib.axes or None
        Define the position of the plot in a figure\n
        Default is `None`\n
    color : str
        Color of the bar\n
        Default is `#325fa8`\n

    Returns
    -------
    None
        Barplot of the distance between the peaks and the TSS of the associated gene(s)

    """
    res = {
        "<-500": [0],
        "-500:-50": [0],
        "-50:-5": [0],
        "-5:0": [0],
        "0:5": [0],
        "5:50": [0],
        "50:500": [0],
        ">500": [0],
    }
    nb = 0

    dist = great.tl.get_dist_to_tss(test, regdom)
    for i in dist.values():
        for j in i:
            if j < -500000:
                res["<-500"][0] += 1
            elif j < -50000:
                res["-500:-50"][0] += 1
            elif j < -5000:
                res["-50:-5"][0] += 1
            elif j < 0:
                res["-5:0"][0] += 1
            elif j < 5000:
                res["0:5"][0] += 1
            elif j < 50000:
                res["5:50"][0] += 1
            elif j < 500000:
                res["50:500"][0] += 1
            else:
                res[">500"][0] += 1
            nb += 1
    df = pd.DataFrame(res).transpose().rename(columns={0: "count"})
    df["percentage"] = (df["count"] / nb) * 100
    df = df.reset_index(drop=False).rename(columns={"index": "distance"})
    g = sns.barplot(data=df, x="distance", y="percentage", color=color, ax=ax)
    for idx, p in enumerate(g.patches):
        g.annotate(str(df.iloc[idx]["count"]), (p.get_x() + p.get_width() / 2, p.get_height()))
    g.set_xlabel("Distance to TSS (kb)", fontsize=15)
    g.set_ylabel("Genomic region (%)", fontsize=15)
    g.set_title("Binned by absolute distance to TSS", fontsize=20)


def graph_absolute_dist_tss(
    test: str or pd.DataFrame,
    regdom: str or pd.DataFrame,
    ax: matplotlib.axes or None = None,
    color: str = "#325fa8",
) -> None:
    """
    Creation of a barplot of the absolute
    distance between the peaks and the TSS of the associated gene(s).

    Parameters
    ----------
    test : str or pd.DataFrame
        Genomic set of peaks to be tested\n
    regdom : str or pd.DataFrame
        Regulatory domain of all genes in the genome\n
    ax : matplotlib.axes or None
        Define the position of the plot in a figure\n
        Default is `None`\n
    color : str
        Color of the bar\n
        Default is `#325fa8`\n

    Returns
    -------
    None
        Barplot of the absolute distance between the peaks and the TSS of the associated gene(s)

    """
    res = {"0:5": [0], "5:50": [0], "50:500": [0], ">500": [0]}
    nb = 0
    dist = great.tl.get_dist_to_tss(test, regdom)

    for i in dist.values():
        for j in i:
            j = abs(j)
            if j < 5000:
                res["0:5"][0] += 1
            elif j < 50000:
                res["5:50"][0] += 1
            elif j < 500000:
                res["50:500"][0] += 1
            else:
                res[">500"][0] += 1
            nb += 1
    df = pd.DataFrame(res).transpose().rename(columns={0: "count"})
    df["percentage"] = (df["count"] / nb) * 100
    df = df.reset_index(drop=False).rename(columns={"index": "distance"})
    g = sns.barplot(data=df, x="distance", y="percentage", color=color, ax=ax)
    for idx, p in enumerate(g.patches):
        g.annotate(str(df.iloc[idx]["count"]), (p.get_x() + p.get_width() / 2, p.get_height()))
    g.set_xlabel("Absolute distance to TSS (kb)", fontsize=15)
    g.set_ylabel("Genomic region (%)", fontsize=15)
    g.set_title("Binned by absolute distance to TSS", fontsize=20)


def get_all_comparison(
    results: dict,
    out_dir: str = "../data/tests/test_data/output/",
    information_folder: str = "../data/human/",
    good_gene_associations: bool = True,
    disp_scatterplot: bool = True,
    stats: bool = True,
):
    """
    Plot the comparaison between greatpy and Great from some files compute by `great.tl.enrichment_multiple`.

    Parameters
    ----------
    results : dict
        Dictionary of results from `great.tl.enrichment_multiple`\n
    out_dir : str
        Path of the output directory with the results of great webserver. \n
        Default is `../data/tests/test_data/output/`\n
    information_folder : str
        path of the folder with the information files for the tests. \n
        Default is `../data/human/`\n
        The input folder should contains the files : \n
        - `information_folder/assembly_eg_hg38/regulatory_domain.bed` \n
        - `information_folder/assembly_eg_hg38/chr_size.bed`\n
    good_gene_associations : bool
        If True, the function return the number of good gene associations\n
    disp_scatterplot : bool
        If True, the function display the scatterplot of the comparaison\n
    stats : bool
        If True, the function return the statistics of the comparaison\n

    Returns
    -------
    pp : pd.DataFrame
        Dataframe of the number of row lost between before preprocessing and after preprocessing\n
    asso : pd.DataFrame
        DataFrame of the number of good gene associations for each file\n
    stats_df : pd.DataFrame
        DataFrame of the statistics of the comparaison for each file\n

    Example
    -------
    >>> test = [
    ...    '../data/tests/test_data/input/09_ERF.bed', '../data/tests/test_data/input/10_MAX.bed',
    ...    '../data/tests/test_data/input/01_random.bed', '../data/tests/test_data/input/04_ultra_hg38.bed',
    ...    '../data/tests/test_data/input/02_srf_hg38.bed', '../data/tests/test_data/input/08_FOXO3.bed',
    ...    '../data/tests/test_data/input/06_height_snps_hg38.bed'
    ...     ]
    >>> results = great.tl.enrichment_multiple(
    ...    tests = t,
    ...    regdom_file=regdom,
    ...    chr_size_file=size,
    ...    annotation_file="../data/human/ontologies.csv",
    ...    annpath=None,
    ...    binom=True,
    ...    hypergeom=True
    ...    )
    >>> pp,asso,stat = get_all_comparison(results)

    >>> pp
    ...    |    | name   |   before_pp_greatpy_size |   before_pp_great_size |   final_size |   %_of_GO_from_great_lost |
    ...    |---:|:-------|-------------------------:|-----------------------:|-------------:|--------------------------:|
    ...    |  0 | ERF    |                     6014 |                   2410 |         1833 |                     23.94 |
    ...    |  1 | MAX    |                     2996 |                   2395 |         1481 |                     38.16 |
    ...    |  2 | random |                      579 |                    197 |          117 |                     40.61 |
    ...    |  3 | ultra  |                     3265 |                   2175 |         1393 |                     35.95 |
    ...    |  4 | srf    |                     4810 |                   2681 |         1854 |                     30.85 |

    >>> asso
    ...    |    | name   |   number_good_gene_asso |   number_genes_asso_lost |   number_gene_asso_excess |
    ...    |---:|:-------|------------------------:|-------------------------:|--------------------------:|
    ...    |  0 | ERF    |                    1456 |                        0 |                        36 |
    ...    |  1 | MAX    |                     428 |                        0 |                         4 |
    ...    |  2 | random |                      57 |                        0 |                         1 |
    ...    |  3 | ultra  |                     496 |                        0 |                         2 |
    ...    |  4 | srf    |                     923 |                        0 |                         7 |

    >>> stat
    ...    |    | name   |   pearson_binom |   pearson_hypergeom |
    ...    |---:|:-------|----------------:|--------------------:|
    ...    |  0 | ERF    |        0.57644  |            0.609552 |
    ...    |  1 | MAX    |        0.601492 |            0.670499 |
    ...    |  2 | random |        0.240765 |            0.124707 |
    ...    |  3 | ultra  |        0.52949  |            0.675438 |
    ...    |  4 | srf    |        0.631909 |            0.597787 |
    """
    pp = {
        "name": [],
        "before_pp_greatpy_size": [],
        "before_pp_great_size": [],
        "final_size": [],
        "%_of_GO_from_great_lost": [],
    }
    asso = {
        "name": [],
        "number_good_gene_asso": [],
        "number_genes_asso_lost": [],
        "number_gene_asso_excess": [],
    }

    stat_df = {"name": [], "pearson_binom": [], "pearson_hypergeom": []}

    for path in results.keys():
        id = path.split("/")[-1].split("_")[0]
        name = path.split("/")[-1].split("_")[1].split(".")[0]
        i = 0
        great_out = ""
        great_asso = ""

        for out_path in os.listdir(out_dir):
            if out_path.split("_")[0] == id:
                if re.match(".*hg19.*", out_path) is not None:
                    assembly = "hg19"
                else:
                    assembly = "hg38"

                if re.match(".*output.*", out_path) is not None:
                    great_out = out_dir + out_path
                else:
                    great_asso = out_dir + out_path

        test = path
        regdom = f"{information_folder}{assembly}/regulatory_domain.bed"

        if great_out == "" or great_asso == "":
            return False

        pp["name"].append(name)
        enrichment_tot = results[test]
        enrichment_tot = great.tl.set_bonferroni(enrichment_tot, 0.05)
        enrichment_tot = great.tl.set_fdr(enrichment_tot, 0.05)

        great_webserver = pd.read_csv(
            great_out,
            sep="\t",
            comment="#",
            names=[
                "ontologie",
                "term_name",
                "ID",
                "binom_p_value",
                "binom_bonferroni",
                "binom_fdr",
                "hyper_p_value",
                "hyper_bonferroni",
                "hyper_fdr",
            ],
            index_col=False,
            dtype={
                "term_name": "object",
                "ID": "object",
                "binom_p_value": "float64",
                "binom_bonferroni": "float64",
                "binom_fdr": "float64",
                "hyper_p_value": "float64",
                "hyper_bonferroni": "float64",
                "hyper_fdr": "float64",
            },
        )
        great_webserver.rename(columns={"ID": "id"}, inplace=True)
        del great_webserver["ontologie"]
        del great_webserver["term_name"]

        pp["before_pp_greatpy_size"].append(enrichment_tot.shape[0])
        enrichment_tot = enrichment_tot[enrichment_tot.index.isin(list(great_webserver["id"]))]
        pp["final_size"].append(enrichment_tot.shape[0])

        pp["before_pp_great_size"].append(great_webserver.shape[0])
        pp["%_of_GO_from_great_lost"].append(
            round(((great_webserver.shape[0] - enrichment_tot.shape[0]) / great_webserver.shape[0]) * 100, 2)
        )
        great_webserver = great_webserver[great_webserver["id"].isin(list(enrichment_tot.index))]

        great_webserver = great_webserver.sort_values("id")

        if disp_scatterplot or stats:
            binom_greatpy = []
            hyper_greatpy = []
            binom_great = []
            hyper_great = []
            for i in range(enrichment_tot.shape[0]):
                go_id = list(enrichment_tot.index)[i]
                curr_enrichment = enrichment_tot.iloc[i]
                curr_great_webserver = great_webserver.loc[great_webserver["id"] == go_id]
                binom_greatpy.append(float(curr_enrichment["binom_p_value"]))
                hyper_greatpy.append(float(curr_enrichment["hypergeom_p_value"]))
                binom_great.append(float(curr_great_webserver["binom_p_value"]))
                hyper_great.append(float(curr_great_webserver["hyper_p_value"]))
            binom = pd.DataFrame({"binom_greatpy": binom_greatpy, "binom_great": binom_great})
            hyper = pd.DataFrame({"hyper_greatpy": hyper_greatpy, "hyper_great": hyper_great})

            if disp_scatterplot:
                fig = plt.figure(figsize=(10, 5), dpi=80)
                fig.subplots_adjust(hspace=0.4, wspace=0.4)
                ax = fig.add_subplot(2, 2, 1)
                great.pl.scatterplot(binom, colname_x="binom_greatpy", colname_y="binom_great", title=None, ax=ax)
                ax = fig.add_subplot(2, 2, 2)
                great.pl.scatterplot(hyper, colname_x="hyper_greatpy", colname_y="hyper_great", title=None, ax=ax)
                fig.suptitle(f"results for {name}", fontsize=18)
                plt.show()

            if stats:
                stat_df["name"].append(name)
                stat_df["pearson_binom"].append(pearsonr(binom_great, binom_greatpy)[0])
                stat_df["pearson_hypergeom"].append(pearsonr(hyper_great, hyper_greatpy)[0])

        if good_gene_associations:
            gene_asso_great = pd.read_csv(
                great_asso,
                sep="\t",
                comment="#",
                names=["ontologies", "gene"],
                index_col=False,
                dtype={"ontologies": "object", "gene": "object"},
                usecols=["gene"],
            )
            gene_asso_greatpy = great.tl.get_association(
                test=pd.read_csv(
                    test,
                    sep="\t",
                    comment="#",
                    usecols=[0, 1, 2],
                    names=["chr", "chr_start", "chr_end"],
                    dtype={"chr": "object", "chr_start": "int64", "chr_end": "int64"},
                ),
                regdom=pd.read_csv(
                    regdom,
                    sep="\t",
                    comment="#",
                    names=["chr", "chr_start", "chr_end", "name", "tss", "strand"],
                    dtype={
                        "chr": "object",
                        "chr_start": "int64",
                        "chr_end": "int64",
                        "name": "object",
                        "tss": "int64",
                        "strand": "object",
                    },
                ),
            )

            in_in = gene_asso_great[gene_asso_great["gene"].isin(gene_asso_greatpy)].shape[0]
            in_out = [i for i in list(gene_asso_great["gene"]) if i not in gene_asso_greatpy]
            out_in = [i for i in gene_asso_greatpy if i not in list(gene_asso_great["gene"])]

            asso["name"].append(name)
            asso["number_good_gene_asso"].append(str(in_in))
            asso["number_genes_asso_lost"].append(str(len(in_out)))
            asso["number_gene_asso_excess"].append(str(len(out_in)))

    if stats and good_gene_associations:
        return pd.DataFrame(pp), pd.DataFrame(asso), pd.DataFrame(stat_df)

    elif stat_df:
        return pd.DataFrame(pp), pd.DataFrame(stat_df)

    elif good_gene_associations:
        pd.DataFrame(pp), pd.DataFrame(asso)
    else:
        return pd.DataFrame(pp)


def scale_data_5_75(data):
    mind = np.min(data)
    maxd = np.max(data)

    if maxd == mind:
        maxd = maxd + 1
        mind = mind - 1

    drange = maxd - mind
    return (((data - mind) / drange * 0.70) + 0.05) * 100


def plot_enrich(data: pd.DataFrame, n_terms: int = 20, color: str = "cool", save: bool = False):
    """
    Creation of a dotplot of the enrichment
    GO term in the inputs datas

    Parameters
    ----------
    data : pd.DataFrame
        Results for greatpy\n
    n_terms : int
        the number of term to be shown. \n
        Default is `20`\n
    color : str
        The color of the cmap in the plot\n
        Default is `"cool"`\n
    save : bool
        Is the plot should be save\n
        Default is `False`\n

    Returns
    -------
    None
        Dotplot of the enrichment n first GO term in the input data

    """
    # Test data input
    if not isinstance(data, pd.DataFrame):
        raise ValueError("Please input a Pandas Dataframe output by gprofiler.")

    if not np.all([term in data.columns for term in ["p_value", "name", "intersection_size"]]):
        raise TypeError(f"The data frame {data} does not contain enrichment results from gprofiler.")

    data_to_plot = data.iloc[:n_terms, :].copy()
    data_to_plot["go.id"] = data_to_plot.index

    min_pval = data_to_plot["p_value"].min()
    max_pval = data_to_plot["p_value"].max()

    # Scale intersection_size to be between 5 and 75 for plotting
    # Note: this is done as calibration was done for values between 5 and 75
    data_to_plot["scaled.overlap"] = scale_data_5_75(data_to_plot["intersection_size"])

    norm = colors.LogNorm(min_pval, max_pval)
    sm = plt.cm.ScalarMappable(cmap=color, norm=norm)
    sm.set_array([])

    rcParams.update({"font.size": 14, "font.weight": "normal"})

    sns.set(style="whitegrid")

    path = plt.scatter(
        x="recall",
        y="name",
        c="p_value",
        cmap=color,
        norm=colors.LogNorm(min_pval, max_pval),
        data=data_to_plot,
        linewidth=1,
        edgecolor="grey",
        s=[(i + 10) ** 1.5 for i in data_to_plot["scaled.overlap"]],
    )
    ax = plt.gca()
    ax.invert_yaxis()

    ax.set_ylabel("")
    ax.set_xlabel("Gene ratio", fontsize=14, fontweight="normal")
    ax.xaxis.grid(False)
    ax.yaxis.grid(True)

    # Shrink current axis by 20%
    box = ax.get_position()
    ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])

    # Get tick marks for this plot
    # Note: 6 ticks maximum
    min_tick = np.floor(np.log10(min_pval)).astype(int)
    max_tick = np.ceil(np.log10(max_pval)).astype(int)
    tick_step = np.ceil((max_tick - min_tick) / 6).astype(int)

    # Ensure no 0 values
    if tick_step == 0:
        tick_step = 1
        min_tick = max_tick - 1

    ticks_vals = [10**i for i in range(max_tick, min_tick - 1, -tick_step)]
    ticks_labs = ["$10^{" + str(i) + "}$" for i in range(max_tick, min_tick - 1, -tick_step)]

    # Colorbar
    fig = plt.gcf()
    cbaxes = fig.add_axes([0.8, 0.15, 0.03, 0.4])
    cbar = ax.figure.colorbar(sm, ticks=ticks_vals, shrink=0.5, anchor=(0, 0.1), cax=cbaxes)
    cbar.ax.set_yticklabels(ticks_labs)
    cbar.set_label("Adjusted p-value", fontsize=14, fontweight="normal")

    # Size legend
    min_olap = data_to_plot["intersection_size"].min()
    max_olap = data_to_plot["intersection_size"].max()
    olap_range = max_olap - min_olap

    # Note: approximate scaled 5, 25, 50, 75 values are calculated
    #      and then rounded to nearest number divisible by 5
    size_leg_vals = [
        np.ceil(i / 5) * 5
        for i in [min_olap, min_olap + (20 / 70) * olap_range, min_olap + (45 / 70) * olap_range, max_olap]
    ]
    size_leg_scaled_vals = scale_data_5_75(size_leg_vals)

    l1 = plt.scatter([], [], s=(size_leg_scaled_vals[0] + 10) ** 1.5, edgecolors="none", color="black")
    l2 = plt.scatter([], [], s=(size_leg_scaled_vals[1] + 10) ** 1.5, edgecolors="none", color="black")
    l3 = plt.scatter([], [], s=(size_leg_scaled_vals[2] + 10) ** 1.5, edgecolors="none", color="black")
    l4 = plt.scatter([], [], s=(size_leg_scaled_vals[3] + 10) ** 1.5, edgecolors="none", color="black")

    labels = [str(int(i)) for i in size_leg_vals]

    leg = plt.legend(
        [l1, l2, l3, l4],
        labels,
        ncol=1,
        frameon=False,
        fontsize=12,
        handlelength=1,
        loc="center left",
        borderpad=1,
        labelspacing=1.4,
        handletextpad=2,
        title="Gene overlap",
        scatterpoints=1,
        bbox_to_anchor=(-2, 1.5),
        facecolor="black",
    )

    if save:
        plt.savefig("dotplot_save.png", dpi=500)

    plt.show()


def make_bubble_heatmap(
    p_val_df: pd.DataFrame,
    odd_ratio_df: pd.DataFrame,
    na_color: str = "gray",
    tickscolorbar: list = [-2, -1, 0, 1, 2],
    vmin: int or float = -2.5,
    vmax: int or float = 2.5,
    heatmap_grid: list = [2, 4, 0, 2, 2, 1],
    circle_legend_grid: list = [2, 4, 0, 2, 2, 1],
    colorbar_grid: list = [2, 5, 0, 3, 2, 1],
    palette_id: str = "RdBu_r",
    cbar_label: str = "cbar_label",
    ncols: int or None = 8,
    marker: str or None = None,
    **kwargs,
) -> None:
    """
    Generate a dotplot with multiple categories

    Parameters
    ----------
    p_val_df : pd.DataFrame
        DataFrame of the p-values computed by enrichment function\n
    odd_ratio_df : pd.DataFrame
        DataFrame of the odds ratios computed by enrichment function\n
    na_color : str
        Color for NA values\n
        Default is `"gray"`\n
    tickscolorbar : list
        List of ticks should be add on the colorbar\n
        Default is `[-2, -1, 0, 1, 2]`\n
    vmin : int or float
        Minimum value for the colorbar\n
        Default is `-2.5`\n
    vmax : int or float
        Maximum value for the colorbar\n
        Default is `2.5`\n
    heatmap_grid : list
        Grid for the heatmap\n
        Default is `[2, 4, 0, 2, 2, 1]`\n
    circle_legend_grid : list
        Grid for the circle legend\n
        Default is `[2, 4, 0, 2, 2, 1]`\n
    colorbar_grid : list
        Grid for the colorbar\n
        Default is `[2, 5, 0, 3, 2, 1]`\n
    palette_id : str
        Palette id for the colorbar\n
        Default is `"RdBu_r"`\n
    cbar_label : str
        Label for the colorbar\n
        Default is `"cbar_label"`\n
    ncols : int or None
        Number of columns for the colorbar\n
        Default is `8`\n
    marker : str or None
        Marker for the dots\n
        Default is `None`\n
    kwargs
        Additional keyword arguments for the plot

    Returns
    -------
    None
        Bubble heatmap plot for gene ontology
    """
    from matplotlib import gridspec

    sns.set_style("white")

    rc("font", weight="normal")
    # rc("alpha",alpha=0.8)

    quantAmplifier = kwargs.get("quantAmplifier", 1)  # factor for size of bubbles

    # function you received on the color gradient
    pallete = None
    if palette_id == "RdBu_r":
        pallete = plt.cm.RdBu_r
    elif palette_id == "Blues":
        pallete = plt.cm.Blues
    elif palette_id == "Reds":
        pallete = plt.cm.Reds
    elif palette_id == "Greens":
        pallete = plt.cm.Greens
    elif palette_id == "Purples":
        pallete = plt.cm.Purples
    elif palette_id == "YlGnBu":
        pallete = plt.cm.YlGnBu
    elif palette_id == "PuOr":
        pallete = plt.cm.PuOr
    else:
        pass
    scalarmap, colorList = get_specific_color_gradient(pallete, np.array(p_val_df), vmin=vmin, vmax=vmax)

    fig = kwargs.get("fig", None)
    if fig is None:
        plt.clf()
        fig = plt.figure(figsize=(kwargs.get("w", 11), kwargs.get("h", 5)))

    nrows, ncols, rowi, coli, rowspan, colspan = heatmap_grid
    gs = gridspec.GridSpec(nrows, ncols)
    ax = plt.subplot(gs[rowi: rowi + rowspan, coli: coli + colspan])

    ylabelList = []
    i = 0

    # to keep the order of dataframes in the same order as sizes, revert the dataframes
    p_val_df = p_val_df.reindex(index=p_val_df.index[::-1])
    odd_ratio_df = odd_ratio_df.reindex(index=odd_ratio_df.index[::-1])

    min_circle_size = kwargs.get("min_circle_size")
    max_circle_size = kwargs.get("max_circle_size")

    for ri, r in p_val_df.iterrows():
        patList = list(r.values)

        sizeList = []
        if min_circle_size is not None and max_circle_size is not None:
            for si in list(odd_ratio_df.loc[ri]):
                if si < min_circle_size:
                    si = min_circle_size
                scaled = (si - min_circle_size) / (max_circle_size - min_circle_size)
                sizeList.append(scaled * quantAmplifier)
        else:
            sizeList = [(si ** kwargs.get("power", 1.0)) * (quantAmplifier) for si in list(odd_ratio_df.loc[ri])]

        colorList = scalarmap.to_rgba(patList)
        colorList = [ci if not np.isnan(vi) else na_color for ci, vi in zip(colorList, patList)]
        x = list(range(len(patList)))
        y = [i] * len(patList)

        # define hataches, linewidths and alphas
        hatches = [None if np.isnan(pi) else None for pi in patList]
        linewidths = (
            [0.1 if np.isnan(pi) else kwargs.get("sig_line_width", 2.0) for pi in patList]
            if kwargs.get("line_widths", None) is None
            else list(kwargs.get("line_widths").loc[ri])
        )
        alphas = [0.5 if np.isnan(alpha) else 1.0 for alpha in patList]

        for xi, yi, si, ci, hatch_i, lw, alpha in zip(x, y, sizeList, colorList, hatches, linewidths, alphas):
            ax.scatter(
                xi,
                yi,
                s=abs(si) * quantAmplifier,
                marker=marker if marker is not None else ("v" if si < 0 else "^"),
                hatch=hatch_i,
                alpha=alpha,
                edgecolor=kwargs.get("edgecolor", "black"),
                color=ci,
                linewidth=lw,
            )
        ylabelList.append(ri)
        i += 1
    if kwargs.get("grid", True):
        plt.grid(True, linewidth=kwargs.get("grid_linewidth", 1.0))
    plt.title(kwargs.get("heatmap_title", "title"), fontname="Arial", fontweight="normal")
    plt.xlabel(kwargs.get("xlab", "xlab"), fontsize=14, fontname="Arial", fontweight="normal")
    plt.ylabel(kwargs.get("ylab", "ylab"), fontsize=14, fontname="Arial", fontweight="normal")

    plt.yticks(list(range(len(ylabelList))))
    ax.set_yticklabels(
        ylabelList,
        fontsize=kwargs.get("yticks_fontsize", 11),
        color="black",
        ha=kwargs.get("ha_ylabs", "right"),
        fontname="Arial",
        fontweight="normal",
        alpha=1,
    )

    remove_top_n_right_ticks(ax)

    plt.xticks(
        list(range(len(p_val_df.columns))),
        p_val_df.columns,
        fontsize=kwargs.get("xticks_fontsize", 11),
        rotation=kwargs.get("rotation_xlabs", 90),
        color="black",
        ha=kwargs.get("ha_xlabs", "center"),
        fontname="Arial",
        fontweight="normal",
    )

    lh, lt = get_legendHandle_for_second_sanity_check_plot(
        quantAmplifier=quantAmplifier * quantAmplifier,
        marker="^" if marker is None else marker,
        fmt=kwargs.get("fmt_legend", "%.1f"),
        lw=0.5,
        min_circle_size=min_circle_size,
        max_circle_size=max_circle_size,
        values=[tick ** kwargs.get("power", 1.0) for tick in kwargs.get("circle_legend_ticks")],
        labels=kwargs.get("circle_legend_ticks"),
    )
    if kwargs.get("show_circle_legend", True):
        l = plt.legend(
            lh,
            lt,
            bbox_to_anchor=kwargs.get("circle_legend_bbox", (1.8, 1)),
            scatterpoints=1,
            title=kwargs.get("circles_legend_title", "circles_legend_title"),
            ncol=1,
            frameon=False,
        )
        # Add the legend manually to the current Axes.
        ax = plt.gca().add_artist(l)

    # this is to add the circle (significant or not
    if kwargs.get("show_sig_legend", False):
        lh, lt = get_legendHandle_for_second_sanity_check_plot(
            quantAmplifier=quantAmplifier * quantAmplifier,
            marker="^" if marker is None else marker,
            fmt=kwargs.get("sig_fmt_legend", "%s"),
            lw=[kwargs.get("sig_line_width", 2.0), 0.0],
            labels=kwargs.get("sig_legend_ticks", ["Yes", "No"]),
            values=kwargs.get("sig_legend_ticks", ["Yes", "No"]),
            edgecolor=kwargs.get("edgecolor", "black"),
            min_size_default=max(kwargs.get("circle_legend_ticks")) ** kwargs.get("power", 1.0),
        )

        plt.legend(
            lh,
            lt,
            bbox_to_anchor=kwargs.get("sig_legend_bbox", (2.4, 1)),
            title=kwargs.get("sig_legend_title", "significant"),
            ncol=1,
            scatterpoints=1,
            frameon=False,
        )

    if kwargs.get("show_colorbar", True):
        cbar = fig.colorbar(
            scalarmap,
            orientation="horizontal",
            format=kwargs.get("cbar_fmt_ticks", "%.1f"),
            ticks=tickscolorbar,
            pad=0.1,
        )
        cbar.ax.tick_params(labelsize=kwargs.get("colorbar_ticks_labelsize", 12))
        cbar.set_label(cbar_label, fontsize=12, fontweight="normal")
    despine_all()


def get_specific_color_gradient(colormap, inputList, **kwargs) -> tuple:
    vmin = kwargs.get("vmin", "blaq")
    vmax = kwargs.get("vmax", "blaq")
    cm = plt.get_cmap(colormap)
    if vmin == "blaq" or vmax == "blaq":
        if type(inputList) == list:
            cNorm = matplotlib.colors.Normalize(vmin=min(inputList), vmax=max(inputList))
        else:
            cNorm = matplotlib.colors.Normalize(vmin=inputList.min(), vmax=inputList.max())
    else:
        cNorm = matplotlib.colors.Normalize(vmin=vmin, vmax=vmax)
    scalarMap = matplotlib.cm.ScalarMappable(norm=cNorm, cmap=cm)
    scalarMap.set_array(inputList)
    colorList = scalarMap.to_rgba(inputList)
    return scalarMap, colorList


def despine_all() -> None:
    sns.despine(offset=10, trim=True, top=True, right=True, left=True, bottom=True)


def remove_top_n_right_ticks(ax) -> None:
    # Hide the right and top spines
    ax.spines["right"].set_visible(False)
    ax.spines["top"].set_visible(False)
    # Only show ticks on the left and bottom spines
    ax.yaxis.set_ticks_position("left")
    ax.xaxis.set_ticks_position("bottom")


def get_legendHandle_for_second_sanity_check_plot(
    quantAmplifier=None,
    labels=None,
    values=None,
    marker="^",
    fmt="%.1f",
    lw=2.0,
    min_circle_size=None,
    max_circle_size=None,
    min_size_default=1.0,
    edgecolor="black",
    color="grey",
):
    labels = [0.1, 0.5, 1.0, 1.5] if labels is None else labels

    legendHandleList = list()
    labelSize = list()
    for i, quantMem in enumerate(values):
        values[i]

        if min_circle_size is not None and max_circle_size is not None:
            scaled = (quantMem - min_circle_size) / (max_circle_size - min_circle_size)
            quantMem = scaled

        next_size = (quantMem * quantAmplifier) if not isinstance(quantMem, str) else min_size_default * quantAmplifier

        legendHandleList.append(
            plt.scatter(
                [],
                [],
                s=next_size,
                color=color,
                edgecolor=edgecolor,
                linewidths=lw if not isinstance(lw, list) else lw[i],
                alpha=0.9,
                marker=marker,
            )
        )
        labelSize.append(fmt % labels[i])
    return legendHandleList, labelSize


def dotplot_multi_sample(
    test_data: dict,
    n_row: int = 5,
    list_id: list = [],
    fig: matplotlib.figure.Figure or None = None,
    show_term_name: bool = False,
    term_name_nchars: int = 30,
    dot_size_amplifier: int or float = 7,
    palette_id: str or matplotlib.cm = "Reds",
    ylab: str = "GO",
    xlab: str = "",
    label_colorbar: str = "-log(p_hypergeometric)",
    marker: str = "o",
    plot_title: str = "Dotplot of enrichment GO terms",
    line_width: int = 0.1,
    circle_legend: str = "log2(odd ratio)",
    save: bool = False,
    **kwargs,
):
    """
    Dotplot of enrichment GO terms for a given list of example genomic regions.

    Parameters
    ----------
    test_data : dict
        dict of multiple tests output\n
    n_row : int
        Number of rows pick in each dataframe.\n
        Default is `5`\n
    list_id : list
        List of IDs to be plotted. If `list_id == []`, any filter will be applied.\n
        Default is `[]`\n
    fig : matplotlib.figure.Figure or None
        Figure to plot the dotplot.\n
        Default is `None`\n
    show_term_name : bool
        Whether to show the GO term name.\n
        Default is `False`\n
    term_name_nchars : int
        Number of characters to show for the GO term name.\n
        Default is `30`\n
    dot_size_amplifier : int or float
        Amplifier for the dot size.\n
        Default is `7`\n
    palette_id : str or matplotlib.cm
        Color palette for the dotplot.\n
        Default is `Reds`\n
    ylab : str
        Y-axis label.\n
        Default is `"GO"`\n
    xlab : str
        X-axis label.\n
        Default is `""`\n
    label_colorbar : str
        Label for the colorbar.\n
        Default is `"-log(p_hypergeometric)"`\n
    marker : str
        Marker for the dotplot.\n
        Default is `"o"`\n
    plot_title : str
        Plot title.\n
        Default is `"Dotplot of enrichment GO terms"`\n
    line_width : int
        Dot line width.\n
        Default is `0.1`\n
    circle_legend : str
        Legend for the circle.\n
        Default is `"log2(odd ratio)"`\n
    kwargs
        Other parameters to be passed to the make make_bubble_heatmap function.

    Returns
    -------
    None
        Dotplot of enrichment GO terms for the given df\n
    p_val : pandas.DataFrame
        Dataframe of plotted p-values\n
    odds_ratio : pandas.DataFrame
        Dataframe of plotted odds ratios\n
    df : pandas.DataFrame
        Dataframe of all results concatenated\n

    Examples
    --------
    >>> test = ["SRF:Ishikawa,A-673-clone-Asp114,K-562,MCF-7,Hep-G2","MAX:K-562,WA01,HeLa-S3", "BACH1:A-549,GM12878"]
    >>> tmp_df = great.tl.enrichment_multiple(
        tests = test, regdom_file="../data/human/hg38/regulatory_domain.bed",
        chr_size_file="../data/human/hg38/chr_size.bed",
        annotation_file="../data/human/ontologies.csv", binom=True, hypergeom=True,)
    >>> p_val,odd_ratio = great.pl.dotplot_multi_sample(tmp_df)

    >>> p_val
    ...    | id         |       0 |       1 |        2 |
    ...    |:-----------|--------:|--------:|---------:|
    ...    | GO:0051292 | 6.28405 | 1       |  1       |
    ...    | GO:0030261 | 5.51863 | 1       |  1       |
    ...    | GO:0001650 | 5.3019  | 1       |  1       |
    ...    | GO:0090096 | 5.29709 | 1       |  1       |
    ...    | GO:0099637 | 5.29709 | 1       |  1       |
    ...    | GO:0004896 | 1       | 7.98072 |  1       |
    ...    | GO:0038165 | 1       | 7.93887 |  1       |
    ...    | GO:0030883 | 1       | 7.03282 |  1       |
    ...    | GO:0048006 | 1       | 7.03282 |  1       |
    ...    | GO:0004924 | 1       | 7.03282 |  1       |
    ...    | GO:0008137 | 1       | 1       | 14.0412  |
    ...    | GO:0015990 | 1       | 1       | 13.2519  |
    ...    | GO:0006120 | 1       | 1       | 11.7472  |
    ...    | GO:0045277 | 1       | 1       | 10.2356  |
    ...    | GO:0030964 | 1       | 1       |  9.74806 |

    >>> odds_ratio
    ...    | id         |       0 |       1 |       2 |
    ...    |:-----------|--------:|--------:|--------:|
    ...    | GO:0051292 | 4.94165 | 0       | 0       |
    ...    | GO:0030261 | 4.39416 | 0       | 0       |
    ...    | GO:0001650 | 2.54406 | 0       | 0       |
    ...    | GO:0090096 | 7.64209 | 0       | 0       |
    ...    | GO:0099637 | 7.64209 | 0       | 0       |
    ...    | GO:0004896 | 0       | 3.58059 | 0       |
    ...    | GO:0038165 | 0       | 6.00685 | 0       |
    ...    | GO:0030883 | 0       | 5.42189 | 0       |
    ...    | GO:0048006 | 0       | 5.42189 | 0       |
    ...    | GO:0004924 | 0       | 5.42189 | 0       |
    ...    | GO:0008137 | 0       | 0       | 3.76802 |
    ...    | GO:0015990 | 0       | 0       | 6.61172 |
    ...    | GO:0006120 | 0       | 0       | 3.65752 |
    ...    | GO:0045277 | 0       | 0       | 5.44179 |
    ...    | GO:0030964 | 0       | 0       | 7.02675 |

    """
    min_p, min_od, max_p, max_od = 100, 100, 0, 0
    gene_tot = {}
    df = pd.DataFrame()

    for i, name in enumerate(test_data.keys()):
        gene_tot[i] = name
        res = test_data[name].sort_values(by="hypergeom_p_value")
        if len(list_id) > 0:
            res = res[res.index.isin(list_id)]
        res = res.iloc[:n_row]

        res = res.reset_index().rename(columns={"index": "id"})
        res["index"] = [i] * res.shape[0]
        frames = [df, res]
        df = pd.concat(frames)
    df["hypergeom_p_value"] = -np.log(df["hypergeom_p_value"])
    df["hypergeometric_fold_enrichment"] = np.log2(df["hypergeometric_fold_enrichment"])
    dup = df[df.duplicated("id") == True]
    duplicate = dup.drop(
        columns=["go_term", "binom_p_value", "binom_fold_enrichment", "intersection_size", "recall"]
    ).to_dict()
    df = df.drop_duplicates("id")
    p_val = df.pivot("id", "index", "hypergeom_p_value").fillna(1).reindex(df["id"])
    odd_ratio = df.pivot("id", "index", "hypergeometric_fold_enrichment").fillna(0).reindex(df["id"])

    if dup.shape[0] > 0:
        for i, name in enumerate(duplicate["id"].values()):
            if duplicate["index"][i] in p_val.columns:
                p_val[duplicate["index"][i]][name] = duplicate["hypergeom_p_value"][i]
                odd_ratio[duplicate["index"][i]][name] = duplicate["hypergeometric_fold_enrichment"][i]
            else:
                p_val[duplicate["index"][i]] = duplicate["hypergeom_p_value"][i]
                odd_ratio[duplicate["index"][i]] = duplicate["hypergeometric_fold_enrichment"][i]

    p_val.rename(columns=gene_tot, inplace=True)
    odd_ratio.rename(columns=gene_tot, inplace=True)

    if show_term_name:
        go_n = []
        for i in range(df.shape[0]):
            curr = df.iloc[i]
            go_n.append(curr["id"] + " " + curr["go_term"][:term_name_nchars])
        p_val["test"] = go_n
        odd_ratio["test"] = go_n

        p_val = p_val.set_index("test")
        odd_ratio = odd_ratio.set_index("test")

    for i in p_val.columns:
        if min_p > p_val[i].min():
            min_p = p_val[i].min()
        if max_p < p_val[i].max():
            max_p = p_val[i].max()

    for i in odd_ratio.columns:
        if min_od > odd_ratio[i].min():
            min_od = odd_ratio[i].min()
        if max_od < odd_ratio[i].max():
            max_od = odd_ratio[i].max()

    max_p = int(round(max_p))
    max_od = int(round(max_od))
    min_p = int(round(min_p))
    min_od = int(round(min_od))

    if min_od == 0:
        min_od = 1

    plt.rcParams.update({"font.size": 14, "font.weight": "normal"})
    plt.tight_layout(pad=0.1, h_pad=0.1, w_pad=0.1)

    make_bubble_heatmap(
        p_val,
        odd_ratio,
        quantAmplifier=dot_size_amplifier,
        circle_legend_ticks=[i for i in range(round(min_od), round(max_od) + 1, 2)],
        palette_id=palette_id,
        ylab=ylab,
        xlab=xlab,
        tickscolorbar=[int(i) for i in range(round(min_p), round(max_p) + 1, 2)],
        vmin=min_p,
        vmax=max_p,
        cbar_label=label_colorbar,
        marker=marker,
        heatmap_title=plot_title,
        sig_line_width=line_width,
        circles_legend_title=circle_legend,
        cbar_fmt_ticks="%.0f",
        fig=fig,
    )
    if save:
        plt.savefig("multidotplot.png", dpi=300)

    return p_val, odd_ratio, df



from math import exp, fabs, lgamma, log

# import bindome as bd
import dask.dataframe as dd
import numpy as np
import pandas as pd
from scipy.special import comb
from statsmodels.stats.multitest import multipletests

pd.options.display.float_format = "{:12.5e}".format


def loader(
    test_data: None or str or pd.DataFrame,
    regdom_file: None or str or pd.DataFrame,
    chr_size_file: None or str or pd.DataFrame,
    annotation_file: None or str or pd.DataFrame,
):
    """
    Load all datasets needed for the enrichment calculation

    Parameters
    ----------
    test_data : None or str or pd.DataFrame
        Genomic set of peaks to be tested\n
    regdom_file : None or str or pd.DataFrame
        Regulatory domain of all genes in the genome\n
    chr_size_file : None or str or pd.DataFrame
        Table with the size of each chromosome\n
    annotation_file : None or str or pd.DataFrame
        Table with the annotation of each gene in the genome\n

    Returns
    -------
    test_data : pd.DataFrame
        Genomic set of peaks to be tested in the good format\n
    regdom : pd.DataFrame
        Regulatory domain of all genes in the genome in the good format\n
    size : pd.DataFrame
        Table with the size of each chromosome in the good format\n
    ann : pd.DataFrame
        Table with the annotation of each gene in the genome in the good format\n

    Examples
    --------
    >>> test,regdom,size,ann = loader(
        "../../data/tests/test_data/input/02_srf_hg38.bed",
        "../../data/human/hg38/regulatory_domain.bed",
        "../../data/human/hg38/chr_size.bed",
        "../data/human/ontologies.csv"
        )

    >>> test.head()
    ...    |    | chr   |   chr_start |   chr_end |
    ...    |---:|:------|------------:|----------:|
    ...    |  0 | chr1  |     1052028 |   1052049 |
    ...    |  1 | chr1  |     1065512 |   1065533 |
    ...    |  2 | chr1  |     1067375 |   1067397 |
    ...    |  3 | chr1  |     1068083 |   1068119 |
    ...    |  4 | chr1  |    10520283 |  10520490 |

    >>> regdom.head()
    ...    |    | chr   |   chr_start |   chr_end | name      |   tss | strand   |
    ...    |---:|:------|------------:|----------:|:----------|------:|:---------|
    ...    |  0 | chr1  |           0 |     22436 | MIR6859-1 | 17436 | -        |
    ...    |  1 | chr1  |       16436 |     22436 | MIR6859-2 | 17436 | -        |
    ...    |  2 | chr1  |       16436 |     22436 | MIR6859-3 | 17436 | -        |
    ...    |  3 | chr1  |       16436 |     28370 | MIR6859-4 | 17436 | -        |
    ...    |  4 | chr1  |       22436 |     34370 | WASH7P    | 29370 | -        |

    >>> size.head()
    ...    |    | chrom   |      size |
    ...    |---:|:--------|----------:|
    ...    |  0 | chr1    | 248956422 |
    ...    |  1 | chr2    | 242193529 |
    ...    |  2 | chr3    | 198295559 |
    ...    |  3 | chr4    | 190214555 |
    ...    |  4 | chr5    | 181538259 |

    >>> ann.head()
    ...    |    | id         | name                                                   | symbol        |
    ...    |---:|:-----------|:-------------------------------------------------------|:--------------|
    ...    |  0 | GO:0003924 | GTPase activity                                        | DNAJC25-GNG10 |
    ...    |  1 | GO:0007186 | G protein-coupled receptor signaling pathway           | DNAJC25-GNG10 |
    ...    |  2 | GO:0003723 | RNA binding                                            | NUDT4B        |
    ...    |  3 | GO:0005829 | cytosol                                                | NUDT4B        |
    ...    |  4 | GO:0008486 | diphosphoinositol-polyphosphate diphosphatase activity | NUDT4B        |

    """

    if type(regdom_file) == str:
        regdom = pd.read_csv(
            regdom_file,
            sep="\t",
            comment="#",
            names=["chr", "chr_start", "chr_end", "name", "tss", "strand"],
            dtype={
                "chr": "object",
                "chr_start": "int64",
                "chr_end": "int64",
                "name": "object",
                "tss": "int64",
                "strand": "object",
            },
        )
    elif type(regdom_file) == pd.DataFrame:
        regdom = regdom_file.iloc[:, :6]
        colname = list(regdom.columns)
        try:
            regdom = regdom.rename(
                columns={
                    colname[0]: "chr",
                    colname[1]: "chr_start",
                    colname[2]: "chr_end",
                    colname[3]: "name",
                    colname[4]: "tss",
                    colname[5]: "strand",
                }
            )
        except:
            print("Error in the format of the regdom file")
            print("The regdom file must have the following columns : chr, chr_start, chr_end, name, tss, strand")
            return False
    else:
        regdom = regdom_file

    if type(test_data) == str:
        test_data = pd.read_csv(
            test_data,
            sep="\t",
            comment="#",
            usecols=[0, 1, 2],
            names=["chr", "chr_start", "chr_end"],
            dtype={"chr": "object", "chr_start": "int64", "chr_end": "int64"},
        )
    elif type(test_data) == pd.DataFrame:
        test_data = test_data.iloc[:, :3]
        colname = list(test_data.columns)
        try:
            test_data = test_data.rename(columns={colname[0]: "chr", colname[1]: "chr_start", colname[2]: "chr_end"})
        except:
            print("Error in test dataframe, please check your input")
            print("Columns should be : chr...(type object), start(type int), end(type int)")
            return False
    else:
        pass

    if type(chr_size_file) == str:
        size = pd.read_csv(
            chr_size_file,
            sep="\t",
            comment="#",
            names=["chrom", "size"],
            dtype={"chrom": "object", "size": "int64"},
        )
    elif type(chr_size_file) == pd.DataFrame:
        size = chr_size_file.iloc[:, :2]
        colname = list(size.columns)
        try:
            size = size.rename(columns={colname[0]: "chrom", colname[1]: "size"})
        except:
            print("Error in the format of the chr_size file")
            print("The chr_size file must have the following columns : chrom, size")
            return False
    else:
        size = chr_size_file

    if type(annotation_file) == str:
        dask_df = dd.read_csv(
            annotation_file,
            sep=";",
            comment="#",
            dtype={
                "ensembl": "object",
                "id": "object",
                "name": "object",
                "ontology.group": "object",
                "gene.name": "object",
                "symbol": "object",
            },
            usecols=["id", "name", "symbol"],
            low_memory=False,
        )
        ann = dask_df.compute()
        ann = ann[ann["id"].str.match("^GO.*") == True]
    elif type(annotation_file) == pd.DataFrame:
        ann = annotation_file.iloc[:, :4]
        colname = list(ann.columns)
        try:
            ann = ann.rename(columns={colname[0]: "id", colname[1]: "name", colname[3]: "symbol"})
        except:
            print("Error in the format of the annotation file")
            print("The annotation file must have the following columns : id, name, symbol")
            return False
    else:
        ann = annotation_file

    return test_data, regdom, size, ann


def _enrichment_binom_and_hypergeom(
    test: pd.DataFrame, regdom: pd.DataFrame, size: pd.DataFrame, ann: pd.DataFrame, asso: list
) -> pd.DataFrame:
    """
    Used to compute the enrichment of the test data using the binomial test and the hypergeometric test.

    Parameters
    ----------
    test : pd.DataFrame
        Genomic set of peaks to be tested
    regdom : pd.DataFrame
        Regulatory domain of all genes in the genome
    chr_size :  pd.DataFrame
        Table with the size of each chromosome
    annotation : pd.DataFrame
        Table with the annotation of each gene in the genome
    asso : list
        List of the association between gene from regdom and peaks from test

    Returns
    -------
    pd.DataFrame
        dataframe contains for every GO ID associate with a every associated gene the p-value for the binomial test and the the hypergeometric test

    Examples
    --------
    >>> test,regdom,size,ann = loader(
        "../../data/tests/test_data/input/02_srf_hg38.bed",
        "../../data/human/hg38/regulatory_domain.bed",
        "../../data/human/hg38/chr_size.bed",
        "../data/human/ontologies.csv"
        )
    >>> enrichment = _enrichment_binom_and_hypergeom(
        test = test,
        regdom = regdom,
        size = size,
        ann = ann,
        asso = get_association(test,regdom)
        )

    >>> enrichment.head()
    ...    |            | go_term                                                          |   binom_p_value |   hypergeom_p_value |
    ...    |:-----------|:-----------------------------------------------------------------|----------------:|--------------------:|
    ...    | GO:0045887 | positive regulation of synaptic growth at neuromuscular junction |     5.17744e-13 |          0.0029275  |
    ...    | GO:0044721 | protein import into peroxisome matrix, substrate release         |     4.83812e-10 |          0.0029275  |
    ...    | GO:0036250 | peroxisome transport along microtubule                           |     4.83812e-10 |          0.0029275  |
    ...    | GO:0016561 | protein import into peroxisome matrix, translocation             |     6.31131e-10 |          0.00584656 |
    ...    | GO:0047485 | protein N-terminus binding                                       |     1.2945e-09  |          0.0050377  |
    """
    # Init Great
    res = {}
    hit = {}

    # init Hypergeom
    hypergeom_gene_set = len(asso)  # get the number of genes in the test gene set.
    hypergeom_total_number_gene = regdom.shape[0]  # get the number of genes in the genome.

    # Init binom
    n_binom = test.shape[0]  # get the number of genomic region in the test set
    total_nu = size["size"].sum()  # get the total number of nucleotides in the genome

    ann_red = ann[ann["symbol"].isin(asso)]
    regdom = regdom[
        regdom["name"].isin(list(ann[ann["id"].isin(list(ann_red["id"]))]["symbol"]))
    ]  # reduction of the regdom file by selecting only the genes whose GO ID is owned by a gene of the association
    len_on_chr = len_regdom(regdom)  # get the length of each regulatory domain

    # Compute for all associating gene and for each GO id associated with the gene the probability.
    for name in asso:
        ann_name_gene = ann[ann["symbol"].isin([name])]
        id = ann_name_gene["id"]
        ann_reduce = ann[ann["id"].isin(list(id))]
        tmp = []
        for i in list(id.unique()):
            gene_imply = ann_reduce[ann_reduce["id"].isin([i])]
            K_hypergeom = gene_imply.shape[0]  # get be the number of genes in the genome with annotation
            curr_regdom = regdom.loc[regdom["name"].isin(list(gene_imply["symbol"]))]
            k_hypergeom = curr_regdom.loc[curr_regdom["name"].isin(asso)].shape[
                0
            ]  # get the number of genes in the test gene set with annotation

            if i not in list(hit.keys()):
                hit[i] = number_of_hits(
                    test, curr_regdom
                )  # get the number of test genomic regions in the regulatory domain of a gene with annotation
            k_binom = hit[i]
            nb_binom = sum(
                len_on_chr[i] for i in curr_regdom["name"]
            )  # get the portion of the genome in the regulatory domain of a gene with annotation
            tmp.append((k_binom, nb_binom, i, gene_imply.iloc[0]["name"], K_hypergeom, k_hypergeom))

        res.update(
            {
                elem[2]: [
                    elem[3],
                    get_binom_pval(n_binom, elem[0], elem[1] / total_nu),
                    elem[0] / (elem[1] / total_nu),  # binom enrichment
                    hypergeom_cdf(hypergeom_total_number_gene, elem[4], hypergeom_gene_set, elem[5]),
                    (elem[5] * hypergeom_total_number_gene) / (hypergeom_gene_set * elem[4]),  # Hypergeom enrichment
                    elem[0],
                    elem[0] / elem[4],
                ]
                for elem in tmp
            }
        )

    return (
        pd.DataFrame(res)
        .transpose()
        .rename(
            columns={
                0: "go_term",
                1: "binom_p_value",
                2: "binom_fold_enrichment",
                3: "hypergeom_p_value",
                4: "hypergeometric_fold_enrichment",
                5: "intersection_size",
                6: "recall",
            }
        )
        .replace(0, np.nan)
        .dropna()
        .sort_values(by="binom_p_value")
    )


def _enrichment_binom(
    test: pd.DataFrame, regdom: pd.DataFrame, size: pd.DataFrame, ann: pd.DataFrame, asso: list
) -> pd.DataFrame:
    """
    Used to compute the enrichment of the test data using the binomial test.

    Parameters
    ----------
    test : pd.DataFrame
        Genomic set of peaks to be tested
    regdom : pd.DataFrame
        Regulatory domain of all genes in the genome
    chr_size :  pd.DataFrame
        Table with the size of each chromosome
    annotation : pd.DataFrame
        Table with the annotation of each gene in the genome
    asso : list
        List of the association between gene from regdom and peaks from test

    Returns
    -------
    pd.DataFrame
        dataframe contains for every GO ID associate with a every associated gene the p-value for the binomial test

    Examples
    --------
    >>> test,regdom,size,ann = loader(
        "../../data/tests/test_data/input/02_srf_hg38.bed",
        "../../data/human/hg38/regulatory_domain.bed",
        "../../data/human/hg38/chr_size.bed",
        "../data/human/ontologies.csv"
        )
    >>> enrichment = _enrichment_binom(
        test = test,
        regdom = regdom,
        size = size,
        ann = ann,
        asso = get_association(test,regdom)
        )
    >>> enrichment.head()
    ...    |            | go_term                                                   |   binom_p_value |   binom_fold_enrichment |   intersection_size |   recall |
    ...    |:-----------|:----------------------------------------------------------|----------------:|------------------------:|--------------------:|---------:|
    ...    | GO:0072749 | cellular response to cytochalasin B                       |     2.21968e-12 |                227251   |                   5 |  5       |
    ...    | GO:0051623 | positive regulation of norepinephrine uptake              |     2.21968e-12 |                227251   |                   5 |  5       |
    ...    | GO:0098973 | structural constituent of postsynaptic actin cytoskeleton |     2.1174e-10  |                 91052.6 |                   5 |  1.25    |
    ...    | GO:0097433 | dense body                                                |     6.40085e-10 |                 16061.8 |                   8 |  1.33333 |
    ...    | GO:0032796 | uropod organization                                       |     2.6988e-09  |                 54544.9 |                   5 |  2.5     |
    """
    # Init Great
    res = {}
    hit = {}

    # Init binom
    n_binom = test.shape[0]  # get the number of genomic region in the test set
    total_nu = size["size"].sum()  # get the total number of nucleotides in the genome

    ann_red = ann[ann["symbol"].isin(asso)]
    regdom = regdom[
        regdom["name"].isin(list(ann[ann["id"].isin(list(ann_red["id"]))]["symbol"]))
    ]  # reduction of the regdom file by selecting only the genes whose GO ID is owned by a gene of the association
    len_on_chr = len_regdom(regdom)  # get the length of each regulatory domain

    # Compute for all associating gene and for each GO id associated with the gene the probability.
    for name in asso:
        ann_name_gene = ann[ann["symbol"].isin([name])]
        id = ann_name_gene["id"]
        ann_reduce = ann[ann["id"].isin(list(id))]
        tmp = []
        for i in list(id.unique()):
            gene_imply = ann_reduce[ann_reduce["id"].isin([i])]
            K = gene_imply.shape[0]
            curr_regdom = regdom.loc[regdom["name"].isin(list(gene_imply["symbol"]))]

            if i not in list(hit.keys()):
                hit[i] = number_of_hits(
                    test, curr_regdom
                )  # get the number of test genomic regions in the regulatory domain of a gene with annotation
            k_binom = hit[i]
            nb_binom = sum(
                len_on_chr[i] for i in curr_regdom["name"]
            )  # get the portion of the genome in the regulatory domain of a gene with annotation
            tmp.append(k_binom, nb_binom, i, gene_imply.iloc[0]["name"], K)

        res.update(
            {
                elem[2]: [
                    elem[3],
                    get_binom_pval(n_binom, elem[0], elem[1] / total_nu),
                    elem[0] / (elem[1] / total_nu),  # binom enrichment
                    elem[0],
                    elem[0] / elem[4],
                ]
                for elem in tmp
            }
        )

    return (
        pd.DataFrame(res)
        .transpose()
        .rename(
            columns={
                0: "go_term",
                1: "binom_p_value",
                2: "binom_fold_enrichment",
                3: "intersection_size",
                4: "recall",
            }
        )
        .sort_values(by="binom_p_value")
    )


def _enrichment_hypergeom(test: pd.DataFrame, regdom: pd.DataFrame, ann: pd.DataFrame, asso: list) -> pd.DataFrame:
    """
    Used to compute the enrichment of the test data using the hypergeometric test.

    Parameters
    ----------
    test : pd.DataFrame
        Genomic set of peaks to be tested
    regdom : pd.DataFrame
        Regulatory domain of all genes in the genome
    chr_size :  pd.DataFrame
        Table with the size of each chromosome
    annotation : pd.DataFrame
        Table with the annotation of each gene in the genome
    asso : list
        List of the association between gene from regdom and peaks from test

    Returns
    -------
    pd.DataFrame
        dataframe contains for every GO ID associate with a every associated gene the p-value for the hypergeometric test

    Examples
    --------
    >>> test,regdom,size,ann = loader(
        "../data/tests/test_data/input/03_srf_hg19.bed",
        "../data/human/hg19/regulatory_domain.bed",
        "../data/human/hg19/chr_size.bed",
        "../data/human/ontologies.csv"
        )
    >>> enrichment = _enrichment_hypergeom(
        test = test,
        regdom = regdom,
        ann = ann,
        asso = get_association(test,regdom)
        )
    >>> enrichment.head()
    ...    |            | go_term                                                                                    |   hypergeom_p_value |   hypergeometric_fold_enrichment |   intersection_size |    recall |
    ...    |:-----------|:-------------------------------------------------------------------------------------------|--------------------:|---------------------------------:|--------------------:|----------:|
    ...    | GO:0015629 | actin cytoskeleton                                                                         |         2.25347e-06 |                          2.73071 |                  27 | 0.116883  |
    ...    | GO:1903979 | negative regulation of microglial cell activation                                          |         0.000302551 |                         17.522   |                   3 | 0.75      |
    ...    | GO:1902626 | assembly of large subunit precursor of preribosome                                         |         0.000302551 |                         17.522   |                   3 | 0.75      |
    ...    | GO:0001077 | proximal promoter DNA-binding transcription activator activity, RNA polymerase II-specific |         0.000504006 |                          1.94689 |                  29 | 0.0833333 |
    ...    | GO:0000977 | RNA polymerase II regulatory region sequence-specific DNA binding                          |         0.000511704 |                          2.03154 |                  26 | 0.0869565 |
    """
    # Init Great
    res = {}

    # Init hypergeom
    hypergeom_total_number_gene = regdom.shape[0]  # get the number of genes in the genome
    hypergeom_gene_set = len(asso)  # get the number of genes in the test gene set.

    ann_red = ann[ann["symbol"].isin(asso)]
    regdom = regdom[
        regdom["name"].isin(list(ann[ann["id"].isin(list(ann_red["id"]))]["symbol"]))
    ]  # reduction of the regdom file by selecting only the genes whose GO ID is owned by a gene of the association

    # Compute for all associating gene and for each GO id associated with the gene the probability.
    for name in asso:
        ann_name_gene = ann[ann["symbol"] == name]
        id = ann_name_gene["id"]
        ann_reduce = ann[ann["id"].isin(list(id))]
        tmp = []
        for i in list(id.unique()):
            gene_imply = ann_reduce[ann_reduce["id"] == i]
            K_hypergeom = gene_imply.shape[0]  # get be the number of genes in the genome with annotation
            curr_regdom = regdom.loc[regdom["name"].isin(list(gene_imply["symbol"]))]
            k_hypergeom = curr_regdom.loc[curr_regdom["name"].isin(asso)].shape[
                0
            ]  # get the number of genes in the test gene set with annotation
            tmp.append((i, gene_imply.iloc[0]["name"], K_hypergeom, k_hypergeom))

        res.update(
            {
                elem[0]: [
                    elem[1],
                    hypergeom_cdf(hypergeom_total_number_gene, elem[2], hypergeom_gene_set, elem[3]),
                    (elem[3] * hypergeom_total_number_gene) / (hypergeom_gene_set * elem[2]),  # hypergeom enrichment
                    elem[3],
                    elem[3] / elem[2],
                ]
                for elem in tmp
            }
        )

    return (
        pd.DataFrame(res)
        .transpose()
        .rename(
            columns={
                0: "go_term",
                1: "hypergeom_p_value",
                2: "hypergeometric_fold_enrichment",
                3: "intersection_size",
                4: "recall",
            }
        )
        .replace(0, np.nan)
        .dropna()
        .sort_values(by="hypergeom_p_value")
    )


def enrichment(
    test_file: str or pd.DataFrame,
    regdom_file: str or pd.DataFrame,
    chr_size_file: str or pd.DataFrame,
    annotation_file: str or pd.DataFrame,
    binom=True,
    hypergeom=True,
) -> pd.DataFrame:
    """
    Compute the enrichment GO terms for the test genomic region

    Parameters
    ----------
    test_file : str or pd.DataFrame
        Genomic set of peaks to be tested\n
    regdom_file : str or pd.DataFrame
        Regulatory domain of all genes in the genome\n
    chr_size_file : str or pd.DataFrame
        Table with the size of each chromosome\n
    annotation_file : str or pd.DataFrame
        Table with the annotation of each gene in the genome\n
    binom : bool (default True)
        If True, the binomial test is used.\n
    hypergeom : bool (default True)
        If True, the hypergeometric test is used.\n

    Returns
    -------
    pd.DataFrame
        dataframe contains for every GO ID associate with a every associated gene the p-value for the hypergeometric test

    Examples
    --------
    >>> test,regdom,size,ann = loader(
        "../data/tests/test_data/input/03_srf_hg19.bed",
        "../data/human/hg19/regulatory_domain.bed",
        "../data/human/hg19/chr_size.bed",
        "../data/human/ontologies.csv"
        )
    >>> enrichment = enrichment(
        test = test,
        regdom = regdom,
        chr_size_file = size,
        ann = ann,
        binom=True,
        hypergeom=True
        )
    >>> enrichment.head()
    ...    |            | go_term                                                   |   binom_p_value |   binom_fold_enrichment |   hypergeom_p_value |   hypergeometric_fold_enrichment |   intersection_size |   recall |
    ...    |:-----------|:----------------------------------------------------------|----------------:|------------------------:|--------------------:|---------------------------------:|--------------------:|---------:|
    ...    | GO:0072749 | cellular response to cytochalasin B                       |     2.21968e-12 |                227251   |          0.0428032  |                         23.3627  |                   5 |  5       |
    ...    | GO:0051623 | positive regulation of norepinephrine uptake              |     2.21968e-12 |                227251   |          0.0428032  |                         23.3627  |                   5 |  5       |
    ...    | GO:0098973 | structural constituent of postsynaptic actin cytoskeleton |     2.1174e-10  |                 91052.6 |          0.160543   |                          5.84068 |                   5 |  1.25    |
    ...    | GO:0097433 | dense body                                                |     6.40085e-10 |                 16061.8 |          0.00141783 |                         11.6814  |                   8 |  1.33333 |
    ...    | GO:0032796 | uropod organization                                       |     2.6988e-09  |                 54544.9 |          0.00182991 |                         23.3627  |                   5 |  2.5     |

    >>> enrichment = enrichment(
        test = test,
        regdom = regdom,
        ann = ann,
        asso = get_association(test,regdom),
        binom=True,
        hypergeom=False
        )
    ...    |            | go_term                                                   |   binom_p_value |   binom_fold_enrichment |   intersection_size |   recall |
    ...    |:-----------|:----------------------------------------------------------|----------------:|------------------------:|--------------------:|---------:|
    ...    | GO:0072749 | cellular response to cytochalasin B                       |     2.21968e-12 |                227251   |                   5 |  5       |
    ...    | GO:0051623 | positive regulation of norepinephrine uptake              |     2.21968e-12 |                227251   |                   5 |  5       |
    ...    | GO:0098973 | structural constituent of postsynaptic actin cytoskeleton |     2.1174e-10  |                 91052.6 |                   5 |  1.25    |
    ...    | GO:0097433 | dense body                                                |     6.40085e-10 |                 16061.8 |                   8 |  1.33333 |
    ...    | GO:0032796 | uropod organization                                       |     2.6988e-09  |                 54544.9 |                   5 |  2.5     |

    >>> enrichment = enrichment(
        test = test,
        regdom = regdom,
        ann = ann,
        asso = get_association(test,regdom),
        binom=False,
        hypergeom=True
        )
    >>> enrichment.head()
    ...    |            | go_term                                                                                    |   hypergeom_p_value |   hypergeometric_fold_enrichment |   intersection_size |    recall |
    ...    |:-----------|:-------------------------------------------------------------------------------------------|--------------------:|---------------------------------:|--------------------:|----------:|
    ...    | GO:0015629 | actin cytoskeleton                                                                         |         2.25347e-06 |                          2.73071 |                  27 | 0.116883  |
    ...    | GO:1903979 | negative regulation of microglial cell activation                                          |         0.000302551 |                         17.522   |                   3 | 0.75      |
    ...    | GO:1902626 | assembly of large subunit precursor of preribosome                                         |         0.000302551 |                         17.522   |                   3 | 0.75      |
    ...    | GO:0001077 | proximal promoter DNA-binding transcription activator activity, RNA polymerase II-specific |         0.000504006 |                          1.94689 |                  29 | 0.0833333 |
    ...    | GO:0000977 | RNA polymerase II regulatory region sequence-specific DNA binding                          |         0.000511704 |                          2.03154 |                  26 | 0.0869565 |

    """
    if not binom and not hypergeom:
        return False

    test, regdom, size, ann = loader(test_file, regdom_file, chr_size_file, annotation_file)
    asso = get_association(
        test, regdom
    )  # get the name of the regulatory domain associated to each genomic region in the test set

    if binom and hypergeom:
        return _enrichment_binom_and_hypergeom(test, regdom, size, ann, asso)

    elif binom:
        return _enrichment_binom(test, regdom, size, ann, asso)

    else:
        return _enrichment_hypergeom(test, regdom, ann, asso)


# TODO : add bindome when it is available
def enrichment_multiple(
    tests: list,
    regdom_file: str or pd.DataFrame,
    chr_size_file: str or pd.DataFrame,
    annotation_file: str or pd.DataFrame,
    annpath: str or None = "../../annotation/",
    binom: bool = True,
    hypergeom: bool = True,
) -> dict:
    """
    Compute the enrichment of GO term for multiple tests sets using bindome or a list of file path.

    Parameters
    ----------
    tests : list
        List of complete name of data to compute\n
    regdom_file : str or pd.DataFrame
        Regulatory domain of all genes in the genome\n
    chr_size_file : str or pd.DataFrame
        Table with the size of each chromosome\n
    annotation_file : str or pd.DataFrame
        Table with the annotation of each gene in the genome\n
    annpath : str or None, optional
        Path to the annotation files used for bindome computation. \n
        If None, the function doesn't use bindome.\n
        Default is `"../../annotation/"`\n
    binom : bool
        If True, the binomial test is used.\n
        Default is `True`\n
    hypergeom : bool
        If True, the hypergeometric test is used. \n
        Default is `True`\n

    Returns
    -------
    dict
        List of dataframe with the enrichment of each test

    Examples
    --------
    >>> tests = ["MAX:K-562,WA01,HeLa-S3", "BACH1:A-549,GM12878"]
    >>> enrichment = enrichment_multiple(
            tests = tests,
            regdom_file="../data/human/hg38/regulatory_domain.bed",
            chr_size_file="../data/human/hg38/chr_size.bed",
            annotation_file="../data/human/ontologies.csv",
            binom=True,
            hypergeom=True,
        )
    >>> enrichment
    ...    {'MAX': pd.DataFrame,
    ...    'BACH1': pd.DataFrame}

    """
    if not binom and not hypergeom:
        return False

    _, regdom, size, ann = loader(None, regdom_file, chr_size_file, annotation_file)

    # if annpath is not None:
    #     bd.bindome.constants.ANNOTATIONS_DIRECTORY = annpath

    res = {}

    for name in tests:
        # if annpath is not None:
        #     name_TF = name.split(":")[0]
        #     tmp_df = bd.bindome.datasets.REMAP2020.get_remap_peaks(name_TF)
        #     tmp = tmp_df[tmp_df[3] == name].iloc[:, 0:3]
        #     tmp = tmp.rename(columns={"chr": "chr", "start": "chr_start", "end": "chr_end"})
        # else:
        tmp = pd.read_csv(
            name,
            sep="\t",
            comment="#",
            usecols=[0, 1, 2],
            names=["chr", "chr_start", "chr_end"],
            dtype={"chr": "object", "chr_start": "int64", "chr_end": "int64"},
        )

        asso = get_association(
            tmp, regdom
        )  # get the name of the regulatory domain associated to each genomic region in the test set

        if binom and hypergeom:
            enrichment = _enrichment_binom_and_hypergeom(tmp, regdom, size, ann, asso)

        elif binom:
            enrichment = _enrichment_binom(tmp, regdom, size, ann, asso)

        else:
            enrichment = _enrichment_hypergeom(tmp, regdom, ann, asso)

        # if annpath is not None:
        #     res[name_TF] = enrichment
        # else:
        res[name] = enrichment
    return res


def set_bonferroni(enrichment_df: pd.DataFrame, alpha: float = 0.05) -> pd.DataFrame:
    """
    Create new columns in the dataframe with the Bonferroni correction

    Parameters
    ----------
    enrichment_df : pd.DataFrame
        Dataframe with the enrichment computed by enrichment function
    alpha : float
        alpha value for the Bonferroni correction\n
        Default is `0.05`\n

    Returns
    -------
    pd.DataFrame
        dataframe new columns with the Bonferroni correction for each p-value

    Examples
    --------
    >>> test,regdom,size,ann = loader(
        "../data/tests/test_data/input/03_srf_hg19.bed",
        "../data/human/hg19/regulatory_domain.bed",
        "../data/human/hg19/chr_size.bed",
        "../data/human/ontologies.csv"
        )
    >>> enrichment = great.tl.enrichment(
        "../data/tests/test_data/input/03_srf_hg19.bed",
        "../data/human/hg19/regulatory_domain.bed",
        "../data/human/hg19/chr_size.bed",
        "../data/human/ontologies.csv",
        binom=True,
        hypergeom=True
        )
    >>> bonferroni = set_bonferroni(enrichment,alpha=0.05)
    >>> bonferroni.head()
    ...    |            | go_term                                                          |   binom_p_value |   hypergeom_p_value |   binom_bonferroni |   hypergeom_bonferroni |
    ...    |:-----------|:-----------------------------------------------------------------|----------------:|--------------------:|-------------------:|-----------------------:|
    ...    | GO:0045887 | positive regulation of synaptic growth at neuromuscular junction |     5.17744e-13 |          0.0029275  |        3.0754e-10  |                      1 |
    ...    | GO:0044721 | protein import into peroxisome matrix, substrate release         |     4.83812e-10 |          0.0029275  |        2.87384e-07 |                      1 |
    ...    | GO:0036250 | peroxisome transport along microtubule                           |     4.83812e-10 |          0.0029275  |        2.87384e-07 |                      1 |
    ...    | GO:0016561 | protein import into peroxisome matrix, translocation             |     6.31131e-10 |          0.00584656 |        3.74892e-07 |                      1 |
    ...    | GO:0047485 | protein N-terminus binding                                       |     1.2945e-09  |          0.0050377  |        7.68931e-07 |                      1 |

    """
    for col in enrichment_df.columns:
        if col in ["binom_p_value", "hypergeom_p_value"]:
            col_split = col.split("_")
            enrichment_df[f"{col_split[0]}_bonferroni"] = multipletests(
                enrichment_df[col], alpha=alpha, method="bonferroni"
            )[1]
    return enrichment_df


def set_fdr(enrichment_df: pd.DataFrame, alpha: float = 0.05) -> pd.DataFrame:
    """
    Create new columns in the dataframe with the fdr correction

    Parameters
    ----------
    enrichment_df : pd.DataFrame
        Dataframe with the enrichment computed by enrichment function
    alpha : float
        alpha value for the fdr correction\n
        Default is `0.05`\n

    Returns
    -------
    pd.DataFrame
        dataframe new columns with the fdr correction for each p-value

    Examples
    --------
    >>> test,regdom,size,ann = loader(
        "../data/tests/test_data/input/03_srf_hg19.bed",
        "../data/human/hg19/regulatory_domain.bed",
        "../data/human/hg19/chr_size.bed",
        "../data/human/ontologies.csv"
        )
    >>> enrichment = great.tl.enrichment(
        "../data/tests/test_data/input/03_srf_hg19.bed",
        "../data/human/hg19/regulatory_domain.bed",
        "../data/human/hg19/chr_size.bed",
        "../data/human/ontologies.csv",
        binom=True,
        hypergeom=True
        )
    >>> fdr = set_fdr(enrichment,alpha=0.05)
    >>> fdr.head()
    ...    |            | go_term                                                          |   binom_p_value |   hypergeom_p_value |   binom_fdr |   hypergeom_fdr |
    ...    |:-----------|:-----------------------------------------------------------------|----------------:|--------------------:|------------:|----------------:|
    ...    | GO:0045887 | positive regulation of synaptic growth at neuromuscular junction |     5.17744e-13 |          0.0029275  | 3.0754e-10  |       0.0913909 |
    ...    | GO:0044721 | protein import into peroxisome matrix, substrate release         |     4.83812e-10 |          0.0029275  | 9.3723e-08  |       0.0913909 |
    ...    | GO:0036250 | peroxisome transport along microtubule                           |     4.83812e-10 |          0.0029275  | 9.3723e-08  |       0.0913909 |
    ...    | GO:0016561 | protein import into peroxisome matrix, translocation             |     6.31131e-10 |          0.00584656 | 9.3723e-08  |       0.0913909 |
    ...    | GO:0047485 | protein N-terminus binding                                       |     1.2945e-09  |          0.0050377  | 1.53786e-07 |       0.0913909 |

    """
    for col in enrichment_df.columns:
        if col in ["binom_p_value", "hypergeom_p_value"]:
            col_split = col.split("_")
            enrichment_df[f"{col_split[0]}_fdr"] = multipletests(enrichment_df[col], alpha=alpha, method="fdr_bh")[1]
    return enrichment_df


def set_threshold(enrichment_df: pd.DataFrame, colname: str, alpha: int = 0.05) -> pd.DataFrame:
    """
    Delete rows according to the p-value of the column taken as argument. By default the alpha value is 0.05

    Parameters
    ----------
    enrichment_df : pd.DataFrame
        Dataframe with the enrichment computed by enrichment function
    colname : str
        name of the column to take the p-value from\n
    alpha : float
        alpha value used for the threshold filter \n
        Default is `0.05`\n

    Returns
    -------
    pd.DataFrame
        dataframe with the rows deleted according to the p-value threshold

    Examples
    --------
    >>> test,regdom,size,ann = loader(
        "../data/tests/test_data/input/03_srf_hg19.bed",
        "../data/human/hg19/regulatory_domain.bed",
        "../data/human/hg19/chr_size.bed",
        "../data/human/ontologies.csv"
        )
    >>> enrichment = great.tl.enrichment(
        "../data/tests/test_data/input/03_srf_hg19.bed",
        "../data/human/hg19/regulatory_domain.bed",
        "../data/human/hg19/chr_size.bed",
        "../data/human/ontologies.csv",
        binom=True,
        hypergeom=True
        )
    >>> enrichment.shape[0]
    ...    594

    >>> significant = set_threshold(enrichment,colname="binom_p_value",alpha=0.05)
    >>> significant.shape[0]
    ...    310

    """
    if colname in enrichment_df.columns:
        enrichment_df = enrichment_df.loc[enrichment_df[colname] <= alpha]
    return enrichment_df


# Utils function used by Great class
def get_association(test: pd.DataFrame, regdom: pd.DataFrame) -> list:
    """
    Determine the names of genes associated with at least one genomic region

    Parameters
    ----------
    test : pd.DataFrame
        df of the tests pics => columns: ["chr","chr_start","chr_end"]\n
    regdom : pd.DataFrame
        df of the regulatory domains => columns: ["chr"	"chr_start"	"chr_end"	"name"	"tss"	"strand"].\n

    Returns
    -------
    res : list
        list of gene associated with at least with one test peak

    Examples
    --------
    >>> test = pd.DataFrame(
        {
            "chr":["chr1"],
            "chr_start":[1052028],
            "chr_end": [1052049]}
        )
    >>> regdom = pd.DataFrame(
        {
            "chr":["chr1","chr1"],
            "chr_start":[1034992,1079306],
            "chr_end": [1115089,1132016],
            "name":["RNF223","C1orf159"],
            "tss":[1074306,1116089],
            "strand":['-','-']
        })
    >>> get_association(test,regdom)
    ...    ['RNF223']

    """
    res = []
    for i in range(test.shape[0]):
        currTest = test.iloc[i]
        regdom_curr_test = regdom.loc[(regdom["chr"] == currTest["chr"])].sort_values("chr_start")
        regdom_curr_test = regdom_curr_test.loc[
            (
                (regdom_curr_test["chr_start"] <= currTest["chr_start"])
                & (regdom_curr_test["chr_end"] >= currTest["chr_end"])
            )
            | (  # regdom overlap totally test
                (regdom_curr_test["chr_start"] >= currTest["chr_start"])
                & (regdom_curr_test["chr_end"] <= currTest["chr_end"])
            )
            | (  # test overlap totally regdom
                (regdom_curr_test["chr_start"] <= currTest["chr_start"])
                & (regdom_curr_test["chr_end"] <= currTest["chr_end"])
                & (regdom_curr_test["chr_end"] >= currTest["chr_start"])
            )
            | (  # regdom overlap not totally test on left side
                (regdom_curr_test["chr_start"] >= currTest["chr_start"])
                & (regdom_curr_test["chr_end"] >= currTest["chr_end"])
                & (regdom_curr_test["chr_start"] <= currTest["chr_end"])
            )  # regdom overlap not totally test on right side
        ]
        res = res + list(regdom_curr_test["name"])
    return list(dict.fromkeys(res))


def len_regdom(regdom: pd.DataFrame) -> dict:
    """
    Calculate for each gene name the size of the regulatory region in the genome

    Parameters
    ----------
    regdom : pd.DataFrame
        df of the regulatory domains => columns: ["chr"	"chr_start"	"chr_end"	"name"	"tss"	"strand"].\n

    Returns
    -------
    dict
        dictionary in which each key corresponds to a gene name
        from regdom and the value is the size of the regulatory
        region for that gene

    Examples
    --------
    >>> regdom = pd.DataFrame(
        {
            "chr":["chr1","chr1"],
            "chr_start":[1034992,1079306],
            "chr_end": [1115089,1132016],
            "name":["RNF223","C1orf159"],
            "tss":[1074306,1116089],
            "strand":['-','-']
            })
    >>> len_regdom(regdom)
    ...    {'RNF223': 80097, 'C1orf159': 52710}

    """
    test = regdom["chr_end"] - regdom["chr_start"]
    return pd.DataFrame({"len": list(test)}, index=regdom["name"]).to_dict()["len"]


def number_of_hits(test: pd.DataFrame, regdom: pd.DataFrame) -> int:
    """
    Calculate the number of hits from several genomic regions and the file describing the regulatory regions

    Parameters
    ----------
    test : pd.DataFrame
        df of the tests pics => columns: ["chr","chr_start","chr_end"]\n
    regdom : pd.DataFrame
        df of the regulatory domains => columns: ["chr"	"chr_start"	"chr_end"	"name"	"tss"	"strand"].\n

    Returns
    -------
    nb : int
        number of hit

    Examples
    --------
    >>> test = pd.DataFrame(
        {
            "chr":["chr1"],
            "chr_start":[1052028],
            "chr_end": [1052049]}
        )
    >>> regdom = pd.DataFrame(
        {
            "chr":["chr1","chr1"],
            "chr_start":[1034992,1079306],
            "chr_end": [1115089,1132016],
            "name":["RNF223","C1orf159"],
            "tss":[1074306,1116089],
            "strand":['-','-']
        })
    >>> number_of_hits(test,regdom)
    ...    1

    """
    nb = 0
    regdom = regdom[["chr", "chr_start", "chr_end"]]
    regdom = regdom[regdom["chr"].isin(list(test["chr"]))]
    for i in range(test.shape[0]):
        chrom = test.iat[i, 0]
        start = test.iat[i, 1]
        end = test.iat[i, 2]
        regdom_np = regdom["chr"].to_numpy()
        reg_start = regdom["chr_start"].to_numpy()
        reg_end = regdom["chr_end"].to_numpy()
        Chr_reduce = np.where(regdom_np == chrom)
        reg_start = np.take(reg_start, Chr_reduce, axis=0)[0]
        reg_end = np.take(reg_end, Chr_reduce, axis=0)[0]

        if any((reg_start <= start) & (reg_end >= end)):
            nb += 1
    return nb


def betacf(a, b, x) -> float:
    """Used by betai: Evaluates continued fraction for incomplete beta function"""
    maxit = 10000
    eps = 3.0e-7
    fpmin = 1.0e-30
    qab = a + b
    qap = a + 1
    qam = a - 1
    c = 1
    d = 1 - qab * x / qap
    if fabs(d) < fpmin:
        d = fpmin
    d = 1 / d
    h = d
    for m in range(1, maxit + 1):
        m2 = 2 * m
        aa = m * (b - m) * x / ((qam + m2) * (a + m2))
        d = 1.0 + aa * d
        if fabs(d) < fpmin:
            d = fpmin
        c = 1.0 + aa / c
        if fabs(c) < fpmin:
            c = fpmin
        d = 1.0 / d
        h *= d * c
        aa = -(a + m) * (qab + m) * x / ((a + m2) * (qap + m2))
        d = 1.0 + aa * d
        if fabs(d) < fpmin:
            d = fpmin
        c = 1.0 + aa / c
        if fabs(c) < fpmin:
            c = fpmin
        d = 1.0 / d
        dell = d * c
        h *= dell
        if fabs(dell - 1.0) < eps:
            break
    if m > maxit:
        print("a or b too big, or MAXIT too small in betacf")
        return False
    return h


def betai(a, b, x) -> float:
    """Returns the incomplete beta function Ix(a, b)."""
    if x < 0 or x > 1:
        # print("bad x in routine betai")
        return False
    if x == 0 or x == 1:
        bt = 0.0
    else:
        bt = exp(lgamma(a + b) - lgamma(a) - lgamma(b) + a * log(x) + b * log(1.0 - x))
    if x < (a + 1) / (a + b + 2):
        return bt * betacf(a, b, x) / a
    return 1 - bt * betacf(b, a, 1 - x) / b


def get_binom_pval(n: int, k: int, p: float) -> float:
    """
    Calculate the binomial probability
    of obtaining k in a set of size n and whose probability is p

    Parameters
    ----------
    n : int
        Number of genomic region in the test set\n
    k : int
        Number of test genomic regions in the regulatory domain of a gene with annotation\n
    p : float
        Percentage of genome annotated\n

    Returns
    -------
    float
        binomial probability

    Examples
    --------
    >>> get_binom_pval(100,2,0.2)
    ...    0.9999999947037065

    """
    if k == 0:
        return 1
    else:
        return betai(k, n - k + 1, p)


def hypergeom_pmf(N: int, K: int, n: int, k: int) -> float:
    """
    Calculate the probability mass function for hypergeometric distribution

    Parameters
    ----------
    N : int
        Total number of gene in the genome\n
    K : int
        Number of genes in the genome with annotation\n
    n : int
        Number of gene in the test set\n
    k : int
        Number of genes in the test gene set with annotation\n

    Returns
    -------
    float
        proability mass function

    Examples
    --------
    >>> hypergeom_pmf(100,10,30,1)
    ...    0.11270773995748315

    """
    Achoosex = comb(K, k, exact=True)
    NAchoosenx = comb(N - K, n - k, exact=True)
    Nchoosen = comb(N, n, exact=True)
    return ((Achoosex) * NAchoosenx) / Nchoosen


def hypergeom_cdf(N: int, K: int, n: int, k: int) -> float:
    """
    Calculate the cumulative density funtion for hypergeometric distribution

    Parameters
    ----------
    N : int
        Total number of gene in the genome\n
    K : int
        Number of genes in the genome with annotation\n
    n : int
        Number of gene in the test set\n
    k : int
        Number of genes in the test gene set with annotation\n

    Returns
    -------
    float
        Cumulative density function

    Examples
    --------
    >>> hypergeom_cdf(100,10,30,1)
    ...    0.9770827595419788

    """
    return np.sum([hypergeom_pmf(N, K, n, x) for x in range(k, min(K, n) + 1)])



import pandas as pd

import greatpy as great


def get_nb_asso_per_region(test: str or pd.DataFrame, regdom: str or pd.DataFrame) -> dict:
    """
    Determine number of peaks associated with each gene in the regulatory domain.

    Parameters
    ----------
    test : str or pd.DataFrame
        path of the file with the tests pics => columns: ["chr","chr_start","chr_end"]\n
    regdom : str or pd.DataFrame
        path of the file with the regulatory domains => columns: ["chr"	"chr_start"	"chr_end"	"name"	"tss"	"strand"].\n

    Returns
    -------
    res : dict
        dict with the number of associated genes per genomic region : \n
        - key = associated gene\n
        - value = number of peaks associated with the gene

    Examples
    --------
    >>> test = pd.DataFrame(
        {
            "chr":["chr1"],
            "chr_start":[1052028],
            "chr_end": [1052049]}
        )
    >>> regdom = pd.DataFrame(
        {
            "chr":["chr1","chr1"],
            "chr_start":[1034992,1079306],
            "chr_end": [1115089,1132016],
            "name":["RNF223","C1orf159"],
            "tss":[1074306,1116089],
            "strand":['-','-']
        })
    >>> get_association(test,regdom)
    ...    {'RNF223':2}

    """
    res = {}
    test, regdom, _, _ = great.tl.loader(test, regdom, None, None)
    for i in range(test.shape[0]):
        currTest = test.iloc[i]
        regdom_curr_test = regdom.loc[(regdom["chr"] == currTest["chr"])].sort_values("chr_start")
        regdom_curr_test = regdom_curr_test.loc[
            (
                (regdom_curr_test["chr_start"] <= currTest["chr_start"])
                & (regdom_curr_test["chr_end"] >= currTest["chr_end"])
            )
            | (  # regdom overlap totally test
                (regdom_curr_test["chr_start"] >= currTest["chr_start"])
                & (regdom_curr_test["chr_end"] <= currTest["chr_end"])
            )
            | (  # test overlap totally regdom
                (regdom_curr_test["chr_start"] <= currTest["chr_start"])
                & (regdom_curr_test["chr_end"] <= currTest["chr_end"])
                & (regdom_curr_test["chr_end"] >= currTest["chr_start"])
            )
            | (  # regdom overlap not totally test on left side
                (regdom_curr_test["chr_start"] >= currTest["chr_start"])
                & (regdom_curr_test["chr_end"] >= currTest["chr_end"])
                & (regdom_curr_test["chr_start"] <= currTest["chr_end"])
            )  # regdom overlap not totally test on right side
        ]
        res[i] = regdom_curr_test.shape[0]
    return res


def get_dist_to_tss(test: str or pd.DataFrame, regdom: str or pd.DataFrame) -> dict:
    """
    Determine the distance from peaks to the transcription start site of the associated gene

    Parameters
    ----------
    test : str or pd.DataFrame
        path of the file with the tests pics => columns: ["chr","chr_start","chr_end"]\n
    regdom : str or pd.DataFrame
        path of the file with the regulatory domains => columns: ["chr"	"chr_start"	"chr_end"	"name"	"tss"	"strand"].\n

    Returns
    -------
    res : dict
        dict with the distance from tss to the associated genes : \n
        - key = number of the input \n
        - value = distance from peaks to tss of associated genes\n

    Examples
    --------
    >>> test = pd.DataFrame(
        {
            "chr":["chr1"],
            "chr_start":[1052028],
            "chr_end": [1052049]}
        )
    >>> regdom = pd.DataFrame(
        {
            "chr":["chr1","chr1"],
            "chr_start":[1034992,1079306],
            "chr_end": [1115089,1132016],
            "name":["RNF223","C1orf159"],
            "tss":[1074306,1116089],
            "strand":['-','-']
        })
    >>> get_association(test,regdom)
    ...    {'RNF223':[-22278]}

    """
    res = {}
    test, regdom, _, _ = great.tl.loader(test, regdom, None, None)
    for i in range(test.shape[0]):
        currTest = test.iloc[i]
        mean_pos_test = (currTest["chr_end"] + currTest["chr_start"]) / 2
        regdom_curr_test = regdom.loc[(regdom["chr"] == currTest["chr"])].sort_values("chr_start")
        regdom_curr_test = regdom_curr_test.loc[
            (
                (regdom_curr_test["chr_start"] <= currTest["chr_start"])
                & (regdom_curr_test["chr_end"] >= currTest["chr_end"])
            )
            | (  # regdom overlap totally test
                (regdom_curr_test["chr_start"] >= currTest["chr_start"])
                & (regdom_curr_test["chr_end"] <= currTest["chr_end"])
            )
            | (  # test overlap totally regdom
                (regdom_curr_test["chr_start"] <= currTest["chr_start"])
                & (regdom_curr_test["chr_end"] <= currTest["chr_end"])
                & (regdom_curr_test["chr_end"] >= currTest["chr_start"])
            )
            | (  # regdom overlap not totally test on left side
                (regdom_curr_test["chr_start"] >= currTest["chr_start"])
                & (regdom_curr_test["chr_end"] >= currTest["chr_end"])
                & (regdom_curr_test["chr_start"] <= currTest["chr_end"])
            )  # regdom overlap not totally test on right side
        ]
        res[i] = []
        for j in range(regdom_curr_test.shape[0]):
            res[i].append(regdom_curr_test.iloc[j]["tss"] - mean_pos_test)
    return res



from .great import (
    betacf,
    betai,
    enrichment,
    enrichment_multiple,
    get_association,
    get_binom_pval,
    hypergeom_cdf,
    hypergeom_pmf,
    len_regdom,
    loader,
    number_of_hits,
    set_bonferroni,
    set_fdr,
    set_threshold,
)
from .regdom import create_regdom
from .utils import get_dist_to_tss, get_nb_asso_per_region
